#-------------------------------------------------------------------------
#
#  Copyright (c) 2009, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Evan Patterson
#  Date:   06/25/09
#
#-------------------------------------------------------------------------

""" A single line text widget that supports functionality common to native
    search widgets.
"""

from __future__ import absolute_import

from traits.api import Bool, Property, Str
from ..toolkit import toolkit_object
from ..basic_editor_factory import BasicEditorFactory


class SearchEditor(BasicEditorFactory):
    """ A single line text widget that supports functionality common to native
        search widgets.
    """

    # The editor class to be created:
    klass = Property

    # The descriptive text for the widget
    text = Str("Search")

    # Is user input set on every keystroke?
    auto_set = Bool(True)

    # Is user input set when the Enter key is pressed?
    enter_set = Bool(False)

    # Whether to show a search button on the widget
    search_button = Bool(True)

    # Whether to show a cancel button on the widget
    cancel_button = Bool(False)

    # Fire this event on the object whenever a search should be triggered,
    # regardless of whether the search term changed
    search_event_trait = Str

    def _get_klass(self):
        """ Returns the toolkit-specific editor class to be instantiated.
        """
        return toolkit_object('search_editor:SearchEditor')



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various editors for single-selection enumerations, for the
wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Property

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.drop_editor file.
from traitsui.editors.enum_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from constants \
    import OKColor, ErrorColor

from helper \
    import enum_values_changed, TraitsUIPanel, disconnect, disconnect_no_id
from functools import reduce


# default formatting function (would import from string, but not in Python 3)
capitalize = lambda s: s.capitalize()


#-------------------------------------------------------------------------
#  'BaseEditor' class:
#-------------------------------------------------------------------------

class BaseEditor(Editor):
    """ Base class for enumeration editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Current set of enumeration names:
    names = Property

    # Current mapping from names to values:
    mapping = Property

    # Current inverse mapping from values to names:
    inverse_mapping = Property

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if factory.name != '':
            self._object, self._name, self._value = \
                self.parse_extended_name(factory.name)
            self.values_changed()
            self._object.on_trait_change(self._values_changed,
                                         ' ' + self._name, dispatch='ui')
        else:
            factory.on_trait_change(self.rebuild_editor, 'values_modified',
                                    dispatch='ui')

    #-------------------------------------------------------------------------
    #  Gets the current set of enumeration names:
    #-------------------------------------------------------------------------

    def _get_names(self):
        """ Gets the current set of enumeration names.
        """
        if self._object is None:
            return self.factory._names

        return self._names

    #-------------------------------------------------------------------------
    #  Gets the current mapping:
    #-------------------------------------------------------------------------

    def _get_mapping(self):
        """ Gets the current mapping.
        """
        if self._object is None:
            return self.factory._mapping

        return self._mapping

    #-------------------------------------------------------------------------
    #  Gets the current inverse mapping:
    #-------------------------------------------------------------------------

    def _get_inverse_mapping(self):
        """ Gets the current inverse mapping.
        """
        if self._object is None:
            return self.factory._inverse_mapping

        return self._inverse_mapping

    #-------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Recomputes the cached data based on the underlying enumeration model:
    #-------------------------------------------------------------------------

    def values_changed(self):
        """ Recomputes the cached data based on the underlying enumeration model.
        """
        self._names, self._mapping, self._inverse_mapping = \
            enum_values_changed(self._value())

    #-------------------------------------------------------------------------
    #  Handles the underlying object model's enumeration set being changed:
    #-------------------------------------------------------------------------

    def _values_changed(self):
        """ Handles the underlying object model's enumeration set being changed.
        """
        self.values_changed()
        self.rebuild_editor()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self._object is not None:
            self._object.on_trait_change(self._values_changed,
                                         ' ' + self._name, remove=True)
        else:
            self.factory.on_trait_change(self.rebuild_editor,
                                         'values_modified', remove=True)

        super(BaseEditor, self).dispose()

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(BaseEditor):
    """ Simple style of enumeration editor, which displays a combo box.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super(SimpleEditor, self).init(parent)

        factory = self.factory

        if factory.evaluate is None:
            self.control = control = wx.Choice(parent, -1, wx.Point(0, 0),
                                               wx.Size(-1, -1), self.names)
            wx.EVT_CHOICE(parent, self.control.GetId(), self.update_object)
        else:
            self.control = control = wx.ComboBox(parent, -1, '',
                                                 wx.Point(0, 0), wx.Size(-1, -1), self.names,
                                                 style=wx.CB_DROPDOWN)
            wx.EVT_COMBOBOX(parent, control.GetId(), self.update_object)
            wx.EVT_TEXT_ENTER(parent, control.GetId(),
                              self.update_text_object)
            wx.EVT_KILL_FOCUS(control, self.on_kill_focus)

            if (not factory.is_grid_cell) and factory.auto_set:
                wx.EVT_TEXT(parent, control.GetId(), self.update_text_object)

        self._no_enum_update = 0
        self.set_tooltip()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(self.control,
                   wx.EVT_COMBOBOX, wx.EVT_TEXT_ENTER, wx.EVT_TEXT)

        disconnect_no_id(self.control, wx.EVT_KILL_FOCUS)

        super(SimpleEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Handles the user selecting a new value from the combo box:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user selecting a new value from the combo box.
        """
        self._no_enum_update += 1
        try:
            new_value = self.mapping[event.GetString()]
            if new_value == self.value and self.factory.is_grid_cell:
                # If the enum editor is in a grid cell and the value did not
                # change, we want the enum editor to go away, reverting back to
                # the normal cell appearance. This is for 2 reasons:
                #  1. it looks nicer
                #  2. if the grid id suddenly closed, wx freaks & causes a
                #     segfault

                grid = self.control.Parent.Parent
                grid.EnableEditing(False)
                grid.EnableEditing(True)

            self.value = new_value

        except:
            from traitsui.api import raise_to_debug
            raise_to_debug()

        self._no_enum_update -= 1

    #-------------------------------------------------------------------------
    #  Handles the user typing text into the combo box text entry field:
    #-------------------------------------------------------------------------

    def update_text_object(self, event):
        """ Handles the user typing text into the combo box text entry field.
        """
        if self._no_enum_update == 0:
            value = self.control.GetValue()
            try:
                value = self.mapping[value]
            except:
                try:
                    value = self.factory.evaluate(value)
                except Exception as excp:
                    self.error(excp)
                    return

            self._no_enum_update += 1
            try:
                self.value = value
                self.control.SetBackgroundColour(OKColor)
                self.control.Refresh()
            except:
                pass
            self._no_enum_update -= 1

    #-------------------------------------------------------------------------
    #  Handles the control losing the keyboard focus:
    #-------------------------------------------------------------------------

    def on_kill_focus(self, event):
        """ Handles the control losing the keyboard focus.
        """
        self.update_text_object(event)
        event.Skip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self._no_enum_update == 0:
            if self.factory.evaluate is None:
                try:
                    self.control.SetStringSelection(
                        self.inverse_mapping[self.value])
                except:
                    pass
            else:
                try:
                    self.control.SetValue(self.str_value)
                except:
                    pass

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self.control.SetBackgroundColour(ErrorColor)
        self.control.Refresh()

    #-------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        # Note: This code is unnecessarily complex due to a strange bug in
        # wxWidgets implementation of the wx.Combobox control that has strange
        # behavior when the current text field value is one of the selection
        # values when 'Clear' is called. In this case, even saving and
        # restoring the text field value does not work, so we go to great
        # lengths to detect this case and avoid using 'Clear', but still get
        # the equivalent visual results. Modify this code at your own risk...

        control = self.control
        clear = True
        cur_name = None
        if self.factory.evaluate is not None:
            n = control.GetCount()
            cur_names = [control.GetString(i) for i in range(n)]
            cur_name = control.GetValue()
            if cur_name in self.names:
                clear = False
                include = True
                for i in range(n - 1, -1, -1):
                    if cur_name == cur_names[i]:
                        include = False
                    else:
                        control.Delete(i)
                for name in self.names:
                    if include or (name != cur_name):
                        control.Append(name)
                cur_name = None
            else:
                point = control.GetInsertionPoint()

        if clear:
            control.Clear()
            control.AppendItems(self.names)

        if cur_name is not None:
            self._no_enum_update += 1
            control.SetValue(cur_name)
            control.SetInsertionPoint(point)
            self._no_enum_update -= 1

        self.update_editor()

#-------------------------------------------------------------------------
#  'RadioEditor' class:
#-------------------------------------------------------------------------


class RadioEditor(BaseEditor):
    """ Enumeration editor, used for the "custom" style, that displays radio
        buttons.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super(RadioEditor, self).init(parent)

        # Create a panel to hold all of the radio buttons:
        self.control = TraitsUIPanel(parent, -1)
        self.rebuild_editor()

    #-------------------------------------------------------------------------
    #  Handles the user clicking one of the 'custom' radio buttons:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user clicking one of the custom radio buttons.
        """
        try:
            self.value = event.GetEventObject().value
        except:
            pass

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        value = self.value
        for button in self.control.GetChildren():
            state = (button.value == value)
            button.SetValue(state)
            if state:
                button.SetFocus()

    #-------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        # Clear any existing content:
        panel = self.control
        panel.SetSizer(None)
        panel.DestroyChildren()

        # Get the current trait value:
        cur_name = self.str_value

        # Create a sizer to manage the radio buttons:
        names = self.names
        mapping = self.mapping
        n = len(names)
        cols = self.factory.cols
        rows = (n + cols - 1) / cols
        incr = [n / cols] * cols
        rem = n % cols
        for i in range(cols):
            incr[i] += (rem > i)
        incr[-1] = -(reduce(lambda x, y: x + y, incr[:-1], 0) - 1)
        if cols > 1:
            sizer = wx.GridSizer(0, cols, 2, 4)
        else:
            sizer = wx.BoxSizer(wx.VERTICAL)

        # Add the set of all possible choices:
        style = wx.RB_GROUP
        index = 0
        for i in range(rows):
            for j in range(cols):
                if n > 0:
                    name = label = names[index]
                    label = self.string_value(label, capitalize)
                    control = wx.RadioButton(panel, -1, label, style=style)
                    control.value = mapping[name]
                    style = 0
                    control.SetValue(name == cur_name)
                    wx.EVT_RADIOBUTTON(panel, control.GetId(),
                                       self.update_object)
                    self.set_tooltip(control)
                    index += incr[j]
                    n -= 1
                else:
                    control = wx.RadioButton(panel, -1, '')
                    control.value = ''
                    control.Show(False)
                sizer.Add(control, 0, wx.NORTH, 5)

        # Set-up the layout:
        panel.SetSizerAndFit(sizer)

#-------------------------------------------------------------------------
#  'ListEditor' class:
#-------------------------------------------------------------------------


class ListEditor(BaseEditor):
    """ Enumeration editor, used for the "custom" style, that displays a list
        box.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super(ListEditor, self).init(parent)

        # Create a panel to hold all of the radio buttons:
        self.control = wx.ListBox(parent, -1, wx.Point(0, 0),
                                  wx.Size(-1, -1), self.names,
                                  style=wx.LB_SINGLE | wx.LB_NEEDED_SB)
        wx.EVT_LISTBOX(parent, self.control.GetId(), self.update_object)
        self.set_tooltip()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(self.control, wx.EVT_LISTBOX)

        super(ListEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Handles the user selecting a list box item:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user selecting a list box item.
        """
        if not self._ignore_update:
            value = self.control.GetStringSelection()
            try:
                value = self.mapping[value]
            except:
                try:
                    value = self.factory.evaluate(value)
                except:
                    pass
            try:
                self.value = value
            except:
                pass

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        control = self.control
        try:
            index = control.FindString(self.inverse_mapping[self.value])
            if index >= 0:
                control.SetSelection(index)
        except:
            pass

    #-------------------------------------------------------------------------
    #  Rebuilds the contents of the editor whenever the original factory
    #  object's 'values' trait changes:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        self._ignore_update = True
        self.control.Clear()
        self.control.AppendItems(self.names)
        self._ignore_update = False

        # fixme: Is this line necessary?
        self.update_editor()

### EOF #######################################################################



template = """
<html>

<head>
<title>${path}</title>
<style type="text/css">

body {
    font-family:Arial, Verdana;
    margin:0px;
    padding:0px;
}

table.dirlist {
    margin:0 auto;
    font-size:13px;
    color:#666;
    min-width:960px;
}

table.dirlist tr.r1 {
    background-color:#f4f4f4;
}

table.dirlist td {
    padding:6px 12px;
    margin:0px;
}

table.dirlist td a.link-dir {
    font-weight:bold;
}

table.dirlist td a {
    text-decoration:none;
}

table.dirlist td a:hover {
    text-decoration:underline;
}

table.dirlist tr {
    padding:4px;
    margin:0px;
}

table.dirlist tr:hover {
    background-color:#e9e9e9;
}


</style>

</head>

<body>

<div class="dirlist-container">

<table class="dirlist">

    <thead>
        <tr>
            <th>File/Directory</th>
            <th>Size</th>
            <th>Created Date</th>
        </tr>
    </thead>
    <tbody>
        % for i, entry in enumerate(dirlist):
        <tr class="${entry['type']} r${i%2}">
            <td><a class="link-${entry['type']}" href="${ entry['path'] }">${entry['name']}</a></td>
            <td>${entry['size']}</td>
            <td>${entry['created_time']}</td>
        </tr>
        % endfor
    </tbody>

</table>

</div>

</body>

</html>

"""


""" The interface for pages in a preferences dialog. """


# Enthought library imports.
from traits.api import Interface, Str


class IPreferencesPage(Interface):
    """ The interface for pages in a preferences dialog. """

    # The page's category (e.g. 'General/Appearence'). The empty string means
    # that this is a top-level page.
    category = Str

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = Str

    # The page name (this is what is shown in the preferences dialog).
    name = Str

    def apply(self):
        """ Apply the page's preferences. """

    # fixme: We would like to be able to have the following API so that
    # developers are not forced into using traits UI for their preferences
    # pages, but at the moment I can't work out how to do it!
##     def create_control(self, parent):
##         """ Create the toolkit-specific control that represents the page. """

##     def destroy_control(self, parent):
##         """ Destroy the toolkit-specific control that represents the page. """

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The starting point for performing naming operations. """


# Local imports.
from context import Context


def InitialContext(environment):
    """ Creates an initial context for beginning name resolution. """

    # Get the class name of the factory that will produce the initial context.
    klass_name = environment.get(Context.INITIAL_CONTEXT_FACTORY)
    if klass_name is None:
        raise ValueError("No initial context factory specified")

    # Import the factory class.
    klass = _import_symbol(klass_name)

    # Create the factory.
    factory = klass()

    # Ask the factory for a context implementation instance.
    return factory.get_initial_context(environment)


# fixme: This is the same code as in the Envisage import manager but we don't
# want naming to be dependent on Envisage, so we need some other package
# for useful 'Python' tools etc.
def _import_symbol(symbol_path):
    """ Imports the symbol defined by 'symbol_path'.

    'symbol_path' is a string in the form 'foo.bar.baz' which is turned
    into an import statement 'from foo.bar import baz' (ie. the last
    component of the name is the symbol name, the rest is the package/
    module path to load it from).

    """

    components = symbol_path.split('.')

    module_name = '.'.join(components[:-1])
    symbol_name = components[-1]

    module = __import__(module_name, globals(), locals(), [symbol_name])
    symbol = getattr(module, symbol_name)

    return symbol

#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2007 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

"""
The Envisage service providing the model state for the single
project plugin.

"""

# Standard imports
import logging
import os
import shutil

# Enthought library imports
from envisage.api import IApplication
from apptools.preferences.api import IPreferences
from traits.api import Any, HasTraits, Instance, List


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class ModelService(HasTraits):
    """
    The Envisage service providing the model state for the single
    project plugin.

    """

    ##########################################################################
    # Attributes (Traits)
    ##########################################################################

    ### public 'ModelService' interface ######################################

    # The Envisage application that this service is part of.
    application = Instance(IApplication)

    # The factory to use for creating new projects
    factory = Instance('envisage.ui.single_project.project_factory.'
        'ProjectFactory')

    # The preferences to be exposed through this service.
    preferences = Instance(IPreferences)

    # The currently open project
    project = Instance('envisage.ui.single_project.project.Project')

    # The current selection within the current project.
    selection = List(Any)


    ##########################################################################
    # 'object' interface.
    ##########################################################################

    ### operator methods #####################################################

    def __init__(self, application, factory, **traits):
        """
        Constructor.

        We require a reference to an Envisage application and a project
        factory to create an instance.

        """

        super(ModelService, self).__init__(application = application,
            factory = factory, **traits)

        return


    ##########################################################################
    # 'ModelService' interface.
    ##########################################################################

    ### public interface #####################################################

    def are_projects_files(self):
        """
        Returns True if project instances are saved as files and False if
        they are saved as directories.

        """

        return self.factory.PROJECT_CLASS.PROJECTS_ARE_FILES


    def clean_location(self, location):
        """
        Ensures that there are no existing files or directories at the
        specified location by removing them.  Exceptions are raised if
        there are any errors cleaning out existing files or directories.

        """

        logger.debug('Trying to clean location [%s]', location)

        if os.path.isfile(location):
            os.path.remove(location)
        else:
            shutil.rmtree(location)

        return


    def get_default_path(self):
        """
        Return the default location for projects.

        """

        return self.factory.PROJECT_CLASS.get_default_path(self.application)


    ### trait handlers #######################################################

    def _project_changed(self, old, new):
        """
        Called whenever the current project is changed.

        We hook this to make sure the new project knows it's current and
        the old project knows it's not.

        """

        logger.debug('Detected project change from [%s] to [%s] in '
            'ModelService [%s]', old, new, self)

        if old is not None:
            old.stop()
        self.selection = []
        if new is not None:
            new.start()

        return


    def _selection_changed(self, old, new):
        """
        Called whenever the selection within the project is changed.

        Implemented simply to log the change.

        """

        logger.debug('ModelService [%s] selection changed from [%s] to [%s] ',
             self, old, new)

        return


### EOF ######################################################################




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/18/2004
#
#------------------------------------------------------------------------------

""" Defines the abstract ViewElement class that all trait view template items
    (i.e., View, Group, Item, Include) derive from.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import re

from traits.api import HasPrivateTraits, Trait, Bool
import traits._py2to3 as _py2to3

from .ui_traits import (ATheme, AnObject, DockStyle, EditorStyle, ExportType,
                        HelpId, Image)

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

label_pat = re.compile(r"^(.*)\[(.*)\](.*)$", re.MULTILINE | re.DOTALL)
label_pat2 = re.compile(r"^(.*){(.*)}(.*)$", re.MULTILINE | re.DOTALL)

#-------------------------------------------------------------------------
#  'ViewElement' class (abstract):
#-------------------------------------------------------------------------


class ViewElement(HasPrivateTraits):
    """ An element of a view.
    """

    #-------------------------------------------------------------------------
    #  Replaces any items which have an 'id' with an Include object with the
    #  same 'id', and puts the object with the 'id' into the specified
    #  ViewElements object:
    #-------------------------------------------------------------------------

    def replace_include(self, view_elements):
        """ Searches the current object's **content** attribute for objects that
        have an **id** attribute, and replaces each one with an Include object
        with the same **id** value, and puts the replaced object into the
        specified ViewElements object.

        Parameters
        ----------
        view_elements : ViewElements object
            Object containing Group, Item, and Include objects
        """
        pass  # Normally overridden in a subclass

    #-------------------------------------------------------------------------
    #  Returns whether or not the object is replacable by an Include object:
    #-------------------------------------------------------------------------

    def is_includable(self):
        """ Returns whether the object is replacable by an Include object.
        """
        return False  # Normally overridden in a subclass

#-------------------------------------------------------------------------
#  'DefaultViewElement' class:
#-------------------------------------------------------------------------


class DefaultViewElement(ViewElement):
    """ A view element that can be used as a default value for traits whose
        value is a view element.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The default context object to edit:
    object = AnObject

    # The default editor style to use:
    style = EditorStyle

    # The default dock style to use:
    dock = DockStyle

    # The default notebook tab image to use:
    image = Image

    # The category of elements dragged out of the view:
    export = ExportType

    # Should labels be added to items in a group?
    show_labels = Bool(True)

    # The default theme to use for a contained item:
    item_theme = ATheme

    # The default theme to use for a contained item's label:
    label_theme = ATheme

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# The container trait used by ViewSubElements:
Container = Trait(DefaultViewElement(), ViewElement)

#-------------------------------------------------------------------------
#  'ViewSubElement' class (abstract):
#-------------------------------------------------------------------------


class ViewSubElement(ViewElement):
    """ Abstract class representing elements that can be contained in a view.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The object this ViewSubElement is contained in; must be a ViewElement.
    container = Container

    # External help context identifier:
    help_id = HelpId

    #-------------------------------------------------------------------------
    #  Splits a string at a specified character:
    #-------------------------------------------------------------------------

    def _split(self, name, value, char, finder, assign, result):
        """ Splits a string at a specified character.
        """
        col = finder(value, char)
        if col < 0:
            return value

        items = (value[:col].strip(), value[col + 1:].strip())
        if items[assign] != '':
            setattr(self, name, items[assign])

        return items[result]

    #-------------------------------------------------------------------------
    #  Sets an object trait if a specified option string is found:
    #-------------------------------------------------------------------------

    def _option(self, string, option, name, value):
        """ Sets a object trait if a specified option string is found.
        """
        col = string.find(option)
        if col >= 0:
            string = string[: col] + string[col + len(option):]
            setattr(self, name, value)

        return string

    #-------------------------------------------------------------------------
    #  Parses any of the one character forms of the 'style' trait:
    #-------------------------------------------------------------------------

    def _parse_style(self, value):
        """ Parses any of the one-character forms of the **style** trait.
        """
        value = self._option(value, '$', 'style', 'simple')
        value = self._option(value, '@', 'style', 'custom')
        value = self._option(value, '*', 'style', 'text')
        value = self._option(value, '~', 'style', 'readonly')
        value = self._split('style', value, ';', _py2to3.str_rfind, 1, 0)

        return value

    #-------------------------------------------------------------------------
    #  Parses a '[label]' value from the string definition:
    #-------------------------------------------------------------------------

    def _parse_label(self, value):
        """ Parses a '[label]' value from the string definition.
        """
        match = label_pat.match(value)
        if match is not None:
            self._parsed_label()
        else:
            match = label_pat2.match(value)

        empty = False
        if match is not None:
            self.label = match.group(2).strip()
            empty = (self.label == '')
            value = match.group(1) + match.group(3)

        return (value, empty)

    #-------------------------------------------------------------------------
    #  Handles a label being found in the string definition:
    #-------------------------------------------------------------------------

    def _parsed_label(self):
        """ Handles a label being found in the string definition.
        """
        pass

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of a specified trait value:
    #-------------------------------------------------------------------------

    def _repr_value(self, value, prefix='', suffix='', ignore=''):
        """ Returns a "pretty print" version of a specified Item trait value.
        """
        if value == ignore:
            return ''

        return '%s%s%s' % (prefix, value, suffix)

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of a list of traits:
    #-------------------------------------------------------------------------

    def _repr_options(self, *names):
        """ Returns a 'pretty print' version of a list of traits.
        """
        result = []
        for name in names:
            value = getattr(self, name)
            if value != self.trait(name).default_value_for(self, name):
                result.append((name, repr(value)))

        if len(result) > 0:
            n = max([len(name) for name, value in result])
            return ',\n'.join(['%s = %s' % (name.ljust(n), value)
                               for name, value in result])

        return None

    #-------------------------------------------------------------------------
    #  Indents each line in a specified string by a specified number of spaces:
    #-------------------------------------------------------------------------

    def _indent(self, string, indent='    '):
        """ Indents each line in a specified string by 4 spaces.
        """
        return '\n'.join([indent + s for s in string.split('\n')])



"""

  fs.tests.test_s3fs:  testcases for the S3FS module

These tests are set up to be skipped by default, since they're very slow,
require a valid AWS account, and cost money.  You'll have to set the '__test__'
attribute the True on te TestS3FS class to get them running.

"""

import unittest

from fs.tests import FSTestCases, ThreadingTestCases
from fs.path import *

from six import PY3
try:
    from fs import s3fs
except ImportError:
    raise unittest.SkipTest("s3fs wasn't importable")    
    

class TestS3FS(unittest.TestCase,FSTestCases,ThreadingTestCases):

    #  Disable the tests by default
    __test__ = False

    bucket = "test-s3fs.rfk.id.au"

    def setUp(self):        
        self.fs = s3fs.S3FS(self.bucket)
        for k in self.fs._s3bukt.list():
            self.fs._s3bukt.delete_key(k)

    def tearDown(self):
        self.fs.close()

    def test_concurrent_copydir(self):
        #  makedir() on S3FS is currently not atomic
        pass

    def test_makedir_winner(self):
        #  makedir() on S3FS is currently not atomic
        pass

    def test_multiple_overwrite(self):
        # S3's eventual-consistency seems to be breaking this test
        pass


class TestS3FS_prefix(TestS3FS):

    def setUp(self):
        self.fs = s3fs.S3FS(self.bucket,"/unittest/files")
        for k in self.fs._s3bukt.list():
            self.fs._s3bukt.delete_key(k)

    def tearDown(self):
        self.fs.close()



# Enthought library imports.
from pyface.tasks.api import DockPane

import logging
log = logging.getLogger(__name__)


class FrameworkPane(DockPane):
    #### trait change handlers

    def _task_changed(self):
        log.debug("Task changed to %s in pane %s" % (self.task, self.id))
        if self.control:
            self.control.set_task(self.task)


class FrameworkFixedPane(DockPane):
    def update_dock_features(self, info):
        DockPane.update_dock_features(self, info)
        info.DockFixed()



# Work in Progress - Do not use
from __future__ import with_statement
from fs.base import FS
from fs.expose.serve import packetstream

from collections import defaultdict
import threading
from threading import Lock, RLock
from json import dumps
import Queue as queue
import socket

from six import b


class PacketHandler(threading.Thread):
    
    def __init__(self, transport, prelude_callback=None):
        super(PacketHandler, self).__init__()
        self.transport = transport
        self.encoder = packetstream.JSONFileEncoder(transport)
        self.decoder = packetstream.JSONDecoder(prelude_callback=None)
        
        self.queues = defaultdict(queue.Queue)            
        self._encoder_lock = threading.Lock()
        self._queues_lock = threading.Lock()
        self._call_id_lock = threading.Lock()
        
        self.call_id = 0
                
    def run(self):
        decoder = self.decoder
        read = self.transport.read
        on_packet = self.on_packet
        while True:
            data = read(1024*16)
            if not data:
                print "No data"
                break
            print "data", repr(data)            
            for header, payload in decoder.feed(data):
                print repr(header)
                print repr(payload)
                on_packet(header, payload)
             
    def _new_call_id(self):
        with self._call_id_lock:
            self.call_id += 1
            return self.call_id
             
    def get_thread_queue(self, queue_id=None):
        if queue_id is None:
            queue_id = threading.current_thread().ident
        with self._queues_lock:
            return self.queues[queue_id]
             
    def send_packet(self, header, payload=''):
        call_id = self._new_call_id()                
        queue_id = threading.current_thread().ident        
        client_ref = "%i:%i" % (queue_id, call_id)        
        header['client_ref'] = client_ref
        
        with self._encoder_lock:
            self.encoder.write(header, payload)
            
        return call_id
        
    def get_packet(self, call_id):
        
        if call_id is not None:
            queue_id = threading.current_thread().ident        
            client_ref = "%i:%i" % (queue_id, call_id)
        else:
            client_ref = None
        
        queue = self.get_thread_queue()
        
        while True:        
            header, payload = queue.get()
            print repr(header)
            print repr(payload)
            if client_ref is not None and header.get('client_ref') != client_ref:
                continue
            break                
        
        return header, payload
                
    def on_packet(self, header, payload):
        client_ref = header.get('client_ref', '')
        queue_id, call_id = client_ref.split(':', 1)
        queue_id = int(queue_id)
        #queue_id = header.get('queue_id', '')
        queue = self.get_thread_queue(queue_id)        
        queue.put((header, payload))
     
     
class _SocketFile(object):
    def __init__(self, socket):
        self.socket = socket
        
    def read(self, size):
        try:
            return self.socket.recv(size)
        except:           
            return b('')        
    
    def write(self, data):
        self.socket.sendall(data)
        
    def close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
     

class _RemoteFile(object):
    
    def __init__(self, path, connection):
        self.path = path
        self.connection = connection  

class RemoteFS(FS):
    
    _meta = { 'thead_safe' : True,
              'network' : True,
              'virtual' : False,
              'read_only' : False,
              'unicode_paths' : True,
              }
    
    def __init__(self, addr='', port=3000, username=None, password=None, resource=None, transport=None):
        self.addr = addr
        self.port = port
        self.username = None
        self.password = None
        self.resource = None
        self.transport = transport
        if self.transport is None:
            self.transport = self._open_connection()
        self.packet_handler = PacketHandler(self.transport)  
        self.packet_handler.start()  
        
        self._remote_call('auth',
                          username=username,
                          password=password,
                          resource=resource)
        
    def _open_connection(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.addr, self.port))
        socket_file = _SocketFile(sock)
        socket_file.write(b('pyfs/0.1\n'))
        return socket_file
    
    def _make_call(self, method_name, *args, **kwargs):
        call = dict(type='rpc',
                    method=method_name,
                    args=args,
                    kwargs=kwargs)
        return call        
        
    def _remote_call(self, method_name, *args, **kwargs):
        call = self._make_call(method_name, *args, **kwargs)
        call_id = self.packet_handler.send_packet(call)
        header, payload = self.packet_handler.get_packet(call_id)
        return header, payload
    
    def ping(self, msg):
        call_id = self.packet_handler.send_packet({'type':'rpc', 'method':'ping'}, msg)
        header, payload = self.packet_handler.get_packet(call_id)
        print "PING"
        print header
        print payload
    
    def close(self):
        self.transport.close()
        self.packet_handler.join()
    
    def open(self, path, mode="r", **kwargs):
        pass
    
    def exists(self, path):
        remote = self._remote_call('exists', path)
        return remote.get('response')
        
        

if __name__ == "__main__":
    
    rfs = RemoteFS()    
    rfs.close()
    



#!/usr/bin/env python

"""
Test code for nc_particles

Not very complete

Designed to be run with pytest

"""
# for py2/3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime

import pytest
import netCDF4
from post_gnome import nc_particles


# test the writer...
def test_writer_with_ref_time():
    w = nc_particles.Writer('junk_file.nc', ref_time=datetime.datetime(2010, 2, 3, 0))
    del w
    nc = netCDF4.Dataset('junk_file.nc')
    units = nc.variables['time'].units
    nc.close()
    assert units == "seconds since 2010-02-03T00:00:00"


def test_write_timestep():
    """very simple version"""
    w = nc_particles.Writer('junk_file.nc',
                            ref_time=datetime.datetime(2010, 2, 3, 0),
                            nc_version=4)
    data = {"longitude": [43.2, 43.3, 43.4],
            "latitude": [31.0, 31.2, 31.3],
            "id": [1, 2, 3]
            }
    w.write_timestep(datetime.datetime(2010, 2, 3, 0), data)
    # and another (same data, but whatever..)
    w.write_timestep(datetime.datetime(2010, 2, 3, 0), data)
    del w
    ## read it back in


def test_write_timestep_wrong_size():
    """very simple version"""
    w = nc_particles.Writer('junk_file.nc',
                            nc_version=4)
    data = {"longitude": [43.2, 43.3, 43.4],
            "latitude": [31.0, 31.2, 31.3],
            "id": [1, 2]
            }
    with pytest.raises(ValueError):
        w.write_timestep(datetime.datetime(2010, 2, 3, 0), data)
    del w

# this one fails when run a part of the test suite
#  but runs fine on its own. -- bug in python-netcdf4?
# def test_multi_close():
#     w = nc_particles.Writer('junk_file.nc',
#                             nc_version=4)
#     w.close()
#     w.close()

if __name__ == "__main__":
    test_multi_close()



# Standard library imports.
import os
import wx

# Enthought library imports.
from traits.api import on_trait_change, List, Instance
from envisage.api import Plugin, ExtensionPoint
from envisage.ui.tasks.api import TasksApplication
from pyface.api import FileDialog, YES, OK, CANCEL
from pyface.tasks.api import Task, TaskWindow
from pyface.action.api import Action, ActionItem, Group, Separator
from pyface.tasks.action.api import SMenuBar, SMenu, TaskActionManagerBuilder, SchemaAddition

from omnivore.framework.actions import OpenAction, ExitAction, PreferencesAction, AboutAction
from omnivore.framework.task import FrameworkTask

import logging
log = logging.getLogger(__name__)


class OSXMinimalTask(FrameworkTask):
    @classmethod
    def can_edit(cls, mime):
        return False


class OSXMenuBarPlugin(Plugin):
    """Plugin providing the minimal menu when the Mac application has no
    windows open.
    """

    OSX_MINIMAL_MENU = 'omnivore.osx_minimal_menu'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'omnivore.framework.osx_menu'

    # The plugin's name (suitable for displaying to the user).
    name = 'OSX Menu'

    #### Extension points offered by this plugin ##############################

    minimal_menu_actions = ExtensionPoint(
        List(Instance(SchemaAddition)), id=OSX_MINIMAL_MENU, desc="""
    
    This extension point allows you to contribute schema additions to the menu
    that will appear when no windows are open in an Mac OS X application.
    
        """
    )

    @on_trait_change('application:started')
    def set_common_menu(self):
        if hasattr(wx.MenuBar, "MacSetCommonMenuBar"):
            self.set_common_menu_29()

    def set_common_menu_29(self):
        menubar = SMenuBar(SMenu(Separator(id="NewGroup", separator=False),
                                 Separator(id="NewGroupEnd", separator=False),
                                 Group(OpenAction(), id="OpenGroup"),
                                 Separator(id="OpenGroupEnd", separator=False),
                                 Separator(id="SaveGroupEnd", separator=False),
                                 Group(ExitAction(), id="ExitGroup"),
                                 id='File', name='&File'),
                           SMenu(PreferencesAction(),
                                 id='Edit', name='&Edit'),
                           SMenu(AboutAction(),
                                 id='Help', name='&Help'),
                           )
        app = wx.GetApp()
        # Create a fake task so we can use the menu creation routines
        window = TaskWindow(application=self.application)
        log.debug("OSXMenuBarPlugin: minimal menu extra items: %s" % str(self.minimal_menu_actions))
        task = OSXMinimalTask(menu_bar=menubar, window=window, extra_actions=self.minimal_menu_actions)

        t = TaskActionManagerBuilder(task=task)
        mgr = t.create_menu_bar_manager()
        control = mgr.create_menu_bar(app)
        wx.MenuBar.MacSetCommonMenuBar(control)

        # Prevent wx from exiting when the last window is closed
        app.SetExitOnFrameDelete(False)

    @on_trait_change('application:application_exiting')
    def kill_wx(self):
        app = wx.GetApp()
        app.ExitMainLoop()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Standard library imports.
import os
import tempfile

from cStringIO import StringIO

# Major package imports.
import wx

# Enthought library imports.
from pyface.resource.api import ResourceFactory

from traits.api import Undefined


class PyfaceResourceFactory(ResourceFactory):
    """ The implementation of a shared resource manager. """

    ###########################################################################
    # 'ResourceFactory' toolkit interface.
    ###########################################################################

    def image_from_file(self, filename):
        """ Creates an image from the data in the specified filename. """

        # N.B 'wx.BITMAP_TYPE_ANY' tells wxPython to attempt to autodetect the
        # --- image format.
        return wx.Image(filename, wx.BITMAP_TYPE_ANY)

    def image_from_data(self, data, filename=None):
        """ Creates an image from the specified data. """
        try:
            return wx.ImageFromStream(StringIO(data))
        except:
            # wx.ImageFromStream is only in wx 2.8 or later(?)
            if filename is Undefined:
                return None

        handle = None
        if filename is None:
            # If there is currently no way in wx to create an image from data,
            # we have write it out to a temporary file and then read it back in:
            handle, filename = tempfile.mkstemp()

        # Write it out...
        tf = open(filename, 'wb')
        tf.write(data)
        tf.close()

        # ... and read it back in!  Lovely 8^()
        image = wx.Image(filename, wx.BITMAP_TYPE_ANY)

        # Remove the temporary file.
        if handle is not None:
            os.close(handle)
            os.unlink(filename)

        return image

#### EOF ######################################################################



#----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2011, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#----------------------------------------------------------------------------

""" Defines the concrete implementations of the traits Toolkit interface for
    the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Define the reference to the exported GUIToolkit object:
#-------------------------------------------------------------------------

import toolkit

# Reference to the GUIToolkit object for wxPython
toolkit = toolkit.GUIToolkit()



'''
Created on 25.9.2010

@author: marekp
'''

import sys
import platform
import stat as statinfo

import fs.errors as errors
from fs.path import pathsplit
try:
    # For non-CPython or older CPython versions.
    # Simplejson also comes with C speedup module which
    # is not in standard CPython >=2.6 library.
    import simplejson as json
except ImportError:
    try:
        import json
    except ImportError:
        print "simplejson (http://pypi.python.org/pypi/simplejson/) required"
        raise
    
from .connection import Connection

python3 = int(platform.python_version_tuple()[0]) > 2

if python3:
    from urllib.error import HTTPError
else:
    from urllib2 import HTTPError 
    
class TahoeUtil:
    def __init__(self, webapi):
        self.connection = Connection(webapi)
        
    def createdircap(self):
        return self.connection.post(u'/uri', params={u't': u'mkdir'}).read()
            
    def unlink(self, dircap, path=None):
        path = self.fixwinpath(path, False)
        self.connection.delete(u'/uri/%s%s' % (dircap, path))
    
    def info(self, dircap, path):
        path = self.fixwinpath(path, False)
        meta = json.load(self.connection.get(u'/uri/%s%s' % (dircap, path), {u't': u'json'}))
        return self._info(path, meta)
            
    def fixwinpath(self, path, direction=True):
        '''
            No, Tahoe really does not support file streams...
            This is ugly hack, because it is not Tahoe-specific.
            Should be move to middleware if will be any.
        '''
        if platform.system() != 'Windows':
            return path
        
        if direction and ':' in path:
                path = path.replace(':', '__colon__')
        elif not direction and '__colon__' in path:
                path = path.replace('__colon__', ':')
        return path
    
    def _info(self, path, data):
        if isinstance(data, list):
            type = data[0]
            data = data[1]
        elif isinstance(data, dict):
            type = data['type']
        else:
            raise errors.ResourceInvalidError('Metadata in unknown format!')
        
        if type == 'unknown':
            raise errors.ResourceNotFoundError(path)
        
        info = {'name': unicode(self.fixwinpath(path, True)),
                'type': type,
                'size': data.get('size', 0),
                'ctime': None,
                'uri': data.get('rw_uri', data.get('ro_uri'))}
        if 'metadata' in data:
            info['ctime'] = data['metadata'].get('ctime')
    
        if info['type'] == 'dirnode':
            info['st_mode'] = 0777 |  statinfo.S_IFDIR 
        else:
            info['st_mode'] = 0644

        return info
        
    def list(self, dircap, path=None):
        path = self.fixwinpath(path, False)
        
        data = json.load(self.connection.get(u'/uri/%s%s' % (dircap, path), {u't': u'json'}))     

        if len(data) < 2 or data[0] != 'dirnode':
            raise errors.ResourceInvalidError('Metadata in unknown format!')
        
        data = data[1]['children']
        for i in data.keys():
            x = self._info(i, data[i])
            yield x

    def mkdir(self, dircap, path):
        path = self.fixwinpath(path, False)    
        path = pathsplit(path)
        
        self.connection.post(u"/uri/%s%s" % (dircap, path[0]), data={u't': u'mkdir', u'name': path[1]})
    
    def move(self, dircap, src, dst):
        if src == '/' or dst == '/':
            raise errors.UnsupportedError("Too dangerous operation, aborting")

        src = self.fixwinpath(src, False)
        dst = self.fixwinpath(dst, False)
        
        src_tuple = pathsplit(src)
        dst_tuple = pathsplit(dst)
        
        if src_tuple[0] == dst_tuple[0]:
            # Move inside one directory
            self.connection.post(u"/uri/%s%s" % (dircap, src_tuple[0]), data={u't': u'rename',
                                        u'from_name': src_tuple[1], u'to_name': dst_tuple[1]})
            return

        # Move to different directory. Firstly create link on dst, then remove from src
        try:
            self.info(dircap, dst)
        except errors.ResourceNotFoundError:
            pass
        else:
            self.unlink(dircap, dst)
         
        uri = self.info(dircap, src)['uri']
        self.connection.put(u"/uri/%s%s" % (dircap, dst), data=uri, params={u't': u'uri'})        
        if uri != self.info(dircap, dst)['uri']:
            raise errors.OperationFailedError('Move failed')
        
        self.unlink(dircap, src)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The base class for all context adapters. """


# Enthought library imports.
from traits.api import Any, Dict, Instance, Property, Str

# Local imports.
from context import Context


class ContextAdapter(Context):
    """ The base class for all context adapters. """

    #### 'ContextAdapter' interface ###########################################

    # The object that we are adapting.
    adaptee = Any

    # The context that the object is in.
    context = Instance(Context)

#### EOF ######################################################################



from traits.api import Interface, Instance, Str, Unicode
from pyface.api import ImageResource


class IAbout(Interface):

    about_title = Unicode

    about_version = Unicode

    about_description = Unicode

    about_website = Str

    about_image = Instance(ImageResource)



"""
fs.s3fs
=======

**Currently only avaiable on Python2 due to boto not being available for Python3**

FS subclass accessing files in Amazon S3

This module provides the class 'S3FS', which implements the FS filesystem
interface for objects stored in Amazon Simple Storage Service (S3).

"""

import os
import datetime
import tempfile
from fnmatch import fnmatch
import stat as statinfo

import boto.s3.connection
from boto.s3.prefix import Prefix
from boto.exception import S3ResponseError

from fs.base import *
from fs.path import *
from fs.errors import *
from fs.remote import *
from fs.filelike import LimitBytesFile
from fs import iotools

import six

# Boto is not thread-safe, so we need to use a per-thread S3 connection.
if hasattr(threading,"local"):
    thread_local = threading.local
else:
    class thread_local(object):
        def __init__(self):
            self._map = {}
        def __getattr__(self,attr):
            try:
                return self._map[(threading.currentThread(),attr)]
            except KeyError:
                raise AttributeError, attr
        def __setattr__(self,attr,value):
            self._map[(threading.currentThread(),attr)] = value


class S3FS(FS):
    """A filesystem stored in Amazon S3.

    This class provides the FS interface for files stored in Amazon's Simple
    Storage Service (S3).  It should be instantiated with the name of the
    S3 bucket to use, and optionally a prefix under which the files should
    be stored.

    Local temporary files are used when opening files from this filesystem,
    and any changes are only pushed back into S3 when the files are closed
    or flushed.
    """

    _meta = {'thread_safe': True,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': False,
             'network': True,
             'atomic.move': True,
             'atomic.copy': True,
             'atomic.makedir': True,
             'atomic.rename': False,
             'atomic.setcontents': True
             }

    class meta:
        PATH_MAX = None
        NAME_MAX = None

    def __init__(self, bucket, prefix="", aws_access_key=None, aws_secret_key=None, separator="/", thread_synchronize=True, key_sync_timeout=1):
        """Constructor for S3FS objects.

        S3FS objects require the name of the S3 bucket in which to store
        files, and can optionally be given a prefix under which the files
        should be stored.  The AWS public and private keys may be specified
        as additional arguments; if they are not specified they will be
        read from the two environment variables AWS_ACCESS_KEY_ID and
        AWS_SECRET_ACCESS_KEY.

        The keyword argument 'key_sync_timeout' specifies the maximum
        time in seconds that the filesystem will spend trying to confirm
        that a newly-uploaded S3 key is available for reading.  For no
        timeout set it to zero.  To disable these checks entirely (and
        thus reduce the filesystem's consistency guarantees to those of
        S3's "eventual consistency" model) set it to None.

        By default the path separator is "/", but this can be overridden
        by specifying the keyword 'separator' in the constructor.
        """
        self._bucket_name = bucket
        self._access_keys = (aws_access_key,aws_secret_key)
        self._separator = separator
        self._key_sync_timeout = key_sync_timeout
        # Normalise prefix to this form: path/to/files/
        prefix = normpath(prefix)
        while prefix.startswith(separator):
            prefix = prefix[1:]
        if not prefix.endswith(separator) and prefix != "":
            prefix = prefix + separator
        if isinstance(prefix,unicode):
            prefix = prefix.encode("utf8")
        if aws_access_key is None:
            if "AWS_ACCESS_KEY_ID" not in os.environ:
                raise CreateFailedError("AWS_ACCESS_KEY_ID not set")
        if aws_secret_key is None:
            if "AWS_SECRET_ACCESS_KEY" not in os.environ:
                raise CreateFailedError("AWS_SECRET_ACCESS_KEY not set")
        self._prefix = prefix
        self._tlocal = thread_local()
        super(S3FS, self).__init__(thread_synchronize=thread_synchronize)

    #  Make _s3conn and _s3bukt properties that are created on demand,
    #  since they cannot be stored during pickling.

    def _s3conn(self):
        try:
            (c,ctime) = self._tlocal.s3conn
            if time.time() - ctime > 60:
                raise AttributeError
            return c
        except AttributeError:
            c = boto.s3.connection.S3Connection(*self._access_keys)
            self._tlocal.s3conn = (c,time.time())
            return c
    _s3conn = property(_s3conn)

    def _s3bukt(self):
        try:
            (b,ctime) = self._tlocal.s3bukt
            if time.time() - ctime > 60:
                raise AttributeError
            return b
        except AttributeError:
            try:
                # Validate by listing the bucket if there is no prefix.
                # If there is a prefix, validate by listing only the prefix
                # itself, to avoid errors when an IAM policy has been applied.
                if self._prefix:
                    b = self._s3conn.get_bucket(self._bucket_name, validate=0)
                    b.get_key(self._prefix)
                else:
                    b = self._s3conn.get_bucket(self._bucket_name, validate=1)
            except S3ResponseError, e:
                if "404 Not Found" not in str(e):
                    raise
                b = self._s3conn.create_bucket(self._bucket_name)
            self._tlocal.s3bukt = (b,time.time())
            return b
    _s3bukt = property(_s3bukt)

    def __getstate__(self):
        state = super(S3FS,self).__getstate__()
        del state['_tlocal']
        return state

    def __setstate__(self,state):
        super(S3FS,self).__setstate__(state)
        self._tlocal = thread_local()

    def __repr__(self):
        args = (self.__class__.__name__,self._bucket_name,self._prefix)
        return '<%s: %s:%s>' % args

    __str__ = __repr__

    def _s3path(self,path):
        """Get the absolute path to a file stored in S3."""
        path = relpath(normpath(path))
        path = self._separator.join(iteratepath(path))
        s3path = self._prefix + path
        if s3path and s3path[-1] == self._separator:
            s3path = s3path[:-1]
        if isinstance(s3path,unicode):
            s3path = s3path.encode("utf8")
        return s3path

    def _uns3path(self,s3path,roots3path=None):
        """Get the local path for a file stored in S3.

        This is essentially the opposite of self._s3path().
        """
        if roots3path is None:
            roots3path = self._s3path("")
        i = len(roots3path)
        return s3path[i:]

    def _sync_key(self,k):
        """Synchronise on contents of the given key.

        Since S3 only offers "eventual consistency" of data, it is possible
        to create a key but be unable to read it back straight away.  This
        method works around that limitation by polling the key until it reads
        back the value expected by the given key.

        Note that this could easily fail if the key is modified by another
        program, meaning the content will never be as specified in the given
        key.  This is the reason for the timeout argument to the construtcor.
        """
        timeout = self._key_sync_timeout
        if timeout is None:
            return k
        k2 = self._s3bukt.get_key(k.name)
        t = time.time()
        while k2 is None or k2.etag != k.etag:
            if timeout > 0:
                if t + timeout < time.time():
                    break
            time.sleep(0.1)
            k2 = self._s3bukt.get_key(k.name)
        return k2

    def _sync_set_contents(self,key,contents):
        """Synchronously set the contents of a key."""
        if isinstance(key,basestring):
            key = self._s3bukt.new_key(key)
        if isinstance(contents,basestring):
            key.set_contents_from_string(contents)
        elif hasattr(contents,"md5"):
            hexmd5 = contents.md5
            b64md5 = hexmd5.decode("hex").encode("base64").strip()
            key.set_contents_from_file(contents,md5=(hexmd5,b64md5))
        else:
            try:
                contents.seek(0)
            except (AttributeError,EnvironmentError):
                tf = tempfile.TemporaryFile()
                data = contents.read(524288)
                while data:
                    tf.write(data)
                    data = contents.read(524288)
                tf.seek(0)
                key.set_contents_from_file(tf)
            else:
                key.set_contents_from_file(contents)
        return self._sync_key(key)

    def makepublic(self, path):
        """Mark given path as publicly accessible using HTTP(S)"""
        s3path = self._s3path(path)
        k = self._s3bukt.get_key(s3path)
        k.make_public()

    def getpathurl(self, path, allow_none=False, expires=3600):
        """Returns a url that corresponds to the given path."""
        s3path = self._s3path(path)
        k = self._s3bukt.get_key(s3path)

        # Is there AllUsers group with READ permissions?
        is_public = True in [grant.permission == 'READ' and
                grant.uri == 'http://acs.amazonaws.com/groups/global/AllUsers'
                for grant in k.get_acl().acl.grants]

        url = k.generate_url(expires, force_http=is_public)

        if url == None:
            if not allow_none:
                raise NoPathURLError(path=path)
            return None

        if is_public:
            # Strip time token; it has no sense for public resource
            url = url.split('?')[0]

        return url

    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=64*1024):
        s3path = self._s3path(path)
        if isinstance(data, six.text_type):
            data = data.encode(encoding=encoding, errors=errors)
        self._sync_set_contents(s3path, data)

    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        """Open the named file in the given mode.

        This method downloads the file contents into a local temporary file
        so that it can be worked on efficiently.  Any changes made to the
        file are only sent back to S3 when the file is flushed or closed.
        """
        if self.isdir(path):
            raise ResourceInvalidError(path)
        s3path = self._s3path(path)
        # Truncate the file if requested
        if "w" in mode:
            k = self._sync_set_contents(s3path,"")
        else:
            k = self._s3bukt.get_key(s3path)
        if k is None:
            # Create the file if it's missing
            if "w" not in mode and "a" not in mode:
                raise ResourceNotFoundError(path)
            if not self.isdir(dirname(path)):
                raise ParentDirectoryMissingError(path)
            k = self._sync_set_contents(s3path,"")
        #  Make sure nothing tries to read past end of socket data
        f = LimitBytesFile(k.size,k,"r")
        #  For streaming reads, return the key object directly
        if mode == "r-":
            return f
        #  For everything else, use a RemoteFileBuffer.
        #  This will take care of closing the socket when it's done.
        return RemoteFileBuffer(self,path,mode,f)

    def exists(self,path):
        """Check whether a path exists."""
        s3path = self._s3path(path)
        s3pathD = s3path + self._separator
        # The root directory always exists
        if self._prefix.startswith(s3path):
            return True
        ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
        for k in ks:
            # A regular file
            if _eq_utf8(k.name,s3path):
                return True
            # A directory
            if _eq_utf8(k.name,s3pathD):
                return True
        return False

    def isdir(self,path):
        """Check whether a path exists and is a directory."""
        s3path = self._s3path(path) + self._separator
        # Root is always a directory
        if s3path == "/" or s3path == self._prefix:
            return True
        # Use a list request so that we return true if there are any files
        # in that directory.  This avoids requiring a special file for the
        # the directory itself, which other tools may not create.
        ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
        try:
            iter(ks).next()
        except StopIteration:
            return False
        else:
            return True

    def isfile(self,path):
        """Check whether a path exists and is a regular file."""
        s3path = self._s3path(path)
        # Root is never a file
        if self._prefix.startswith(s3path):
            return False
        k = self._s3bukt.get_key(s3path)
        if k is not None:
            return True
        return False

    def listdir(self,path="./",wildcard=None,full=False,absolute=False,
                               dirs_only=False,files_only=False):
        """List contents of a directory."""
        return list(self.ilistdir(path,wildcard,full,absolute,
                                       dirs_only,files_only))

    def listdirinfo(self,path="./",wildcard=None,full=False,absolute=False,
                                   dirs_only=False,files_only=False):
        return list(self.ilistdirinfo(path,wildcard,full,absolute,
                                           dirs_only,files_only))

    def ilistdir(self,path="./",wildcard=None,full=False,absolute=False,
                                dirs_only=False,files_only=False):
        """List contents of a directory."""
        keys = self._iter_keys(path)
        entries = self._filter_keys(path,keys,wildcard,full,absolute,
                                         dirs_only,files_only)
        return (nm for (nm,k) in entries)

    def ilistdirinfo(self,path="./",wildcard=None,full=False,absolute=False,
                                    dirs_only=False,files_only=False):
        keys = self._iter_keys(path)
        entries = self._filter_keys(path,keys,wildcard,full,absolute,
                                         dirs_only,files_only)
        return ((nm,self._get_key_info(k,nm)) for (nm,k) in entries)

    def _iter_keys(self,path):
        """Iterator over keys contained in the given directory.

        This generator yields (name,key) pairs for each entry in the given
        directory.  If the path is not a directory, it raises the approprate
        error.
        """
        s3path = self._s3path(path) + self._separator
        if s3path == "/":
            s3path = ""
        isDir = False
        for k in self._s3bukt.list(prefix=s3path,delimiter=self._separator):
            if not isDir:
                isDir = True
            # Skip over the entry for the directory itself, if it exists
            name = self._uns3path(k.name,s3path)
            if name != "":
                if not isinstance(name,unicode):
                    name = name.decode("utf8")
                if name.endswith(self._separator):
                    name = name[:-1]
                yield (name,k)
        if not isDir:
            if s3path != self._prefix:
                if self.isfile(path):
                    msg = "that's not a directory: %(path)s"
                    raise ResourceInvalidError(path,msg=msg)
                raise ResourceNotFoundError(path)

    def _key_is_dir(self, k):
        if isinstance(k,Prefix):
            return True
        if k.name.endswith(self._separator):
            return True
        return False

    def _filter_keys(self,path,keys,wildcard,full,absolute,
                               dirs_only,files_only):
        """Filter out keys not matching the given criteria.

        Given a (name,key) iterator as returned by _iter_keys, this method
        applies the given filtering criteria and returns a filtered iterator.
        """
        sep = self._separator
        if dirs_only and files_only:
            raise ValueError("dirs_only and files_only can not both be True")
        if dirs_only:
            keys = ((nm,k) for (nm,k) in keys if self._key_is_dir(k))
        elif files_only:
            keys = ((nm,k) for (nm,k) in keys if not self._key_is_dir(k))
        if wildcard is not None:
            if callable(wildcard):
                keys = ((nm,k) for (nm,k) in keys if wildcard(nm))
            else:
                keys = ((nm,k) for (nm,k) in keys if fnmatch(nm,wildcard))
        if full:
            return ((relpath(pathjoin(path, nm)),k) for (nm,k) in keys)
        elif absolute:
            return ((abspath(pathjoin(path, nm)),k) for (nm,k) in keys)
        return keys

    def makedir(self,path,recursive=False,allow_recreate=False):
        """Create a directory at the given path.

        The 'mode' argument is accepted for compatibility with the standard
        FS interface, but is currently ignored.
        """
        s3path = self._s3path(path)
        s3pathD = s3path + self._separator
        if s3pathD == self._prefix:
            if allow_recreate:
                return
            msg = "Can not create a directory that already exists"\
                  " (try allow_recreate=True): %(path)s"
            raise DestinationExistsError(path, msg=msg)
        s3pathP = self._s3path(dirname(path))
        if s3pathP:
            s3pathP = s3pathP + self._separator
        # Check various preconditions using list of parent dir
        ks = self._s3bukt.list(prefix=s3pathP,delimiter=self._separator)
        if s3pathP == self._prefix:
            parentExists = True
        else:
            parentExists = False
        for k in ks:
            if not parentExists:
                parentExists = True
            if _eq_utf8(k.name,s3path):
                # It's already a file
                msg = "Destination exists as a regular file: %(path)s"
                raise ResourceInvalidError(path, msg=msg)
            if _eq_utf8(k.name,s3pathD):
                # It's already a directory
                if allow_recreate:
                    return
                msg = "Can not create a directory that already exists"\
                      " (try allow_recreate=True): %(path)s"
                raise DestinationExistsError(path, msg=msg)
        # Create parent if required
        if not parentExists:
            if recursive:
                self.makedir(dirname(path),recursive,allow_recreate)
            else:
                msg = "Parent directory does not exist: %(path)s"
                raise ParentDirectoryMissingError(path, msg=msg)
        # Create an empty file representing the directory
        if s3pathD not in ('/', ''):
            self._sync_set_contents(s3pathD,"")

    def remove(self,path):
        """Remove the file at the given path."""
        s3path = self._s3path(path)
        ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
        for k in ks:
            if _eq_utf8(k.name,s3path):
                break
            if _startswith_utf8(k.name,s3path + "/"):
                msg = "that's not a file: %(path)s"
                raise ResourceInvalidError(path,msg=msg)
        else:
            raise ResourceNotFoundError(path)
        self._s3bukt.delete_key(s3path)
        k = self._s3bukt.get_key(s3path)
        while k:
            k = self._s3bukt.get_key(s3path)

    def removedir(self,path,recursive=False,force=False):
        """Remove the directory at the given path."""
        if normpath(path) in ('', '/'):
            raise RemoveRootError(path)
        s3path = self._s3path(path)
        if s3path != self._prefix:
            s3path = s3path + self._separator
        if force:
            #  If we will be forcibly removing any directory contents, we
            #  might as well get the un-delimited list straight away.
            ks = self._s3bukt.list(prefix=s3path)
        else:
            ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
        # Fail if the directory is not empty, or remove them if forced
        found = False
        for k in ks:
            found = True
            if not _eq_utf8(k.name,s3path):
                if not force:
                    raise DirectoryNotEmptyError(path)
                self._s3bukt.delete_key(k.name)
        if not found:
            if self.isfile(path):
                msg = "removedir() called on a regular file: %(path)s"
                raise ResourceInvalidError(path,msg=msg)
            if path not in ("","/"):
                raise ResourceNotFoundError(path)
        self._s3bukt.delete_key(s3path)
        if recursive and path not in ("","/"):
            pdir = dirname(path)
            try:
                self.removedir(pdir,recursive=True,force=False)
            except DirectoryNotEmptyError:
                pass

    def rename(self,src,dst):
        """Rename the file at 'src' to 'dst'."""
        # Actually, in S3 'rename' is exactly the same as 'move'
        if self.isfile(src):
            self.move(src,dst)
        else:
            self.movedir(src,dst)

    def getinfo(self,path):
        s3path = self._s3path(path)
        if path in ("","/"):
            k = Prefix(bucket=self._s3bukt,name="/")
        else:
            k = self._s3bukt.get_key(s3path)
            if k is None:
                ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
                for k in ks:
                    if isinstance(k,Prefix):
                        break
                else:
                    raise ResourceNotFoundError(path)
        return self._get_key_info(k,path)

    def _get_key_info(self,key,name=None):
        info = {}
        if name is not None:
            info["name"] = basename(name)
        else:
            info["name"] = basename(self._uns3key(k.name))
        if self._key_is_dir(key):
            info["st_mode"] = 0700 | statinfo.S_IFDIR
        else:
            info["st_mode"] =  0700 | statinfo.S_IFREG
        if hasattr(key,"size"):
            info['size'] = int(key.size)
        etag = getattr(key,"etag",None)
        if etag is not None:
            if isinstance(etag,unicode):
               etag = etag.encode("utf8")
            info['etag'] = etag.strip('"').strip("'")
        if hasattr(key,"last_modified"):
            # TODO: does S3 use any other formats?
            fmt = "%a, %d %b %Y %H:%M:%S %Z"
            try:
                mtime = datetime.datetime.strptime(key.last_modified,fmt)
                info['modified_time'] = mtime
            except ValueError:
                pass
        return info

    def desc(self,path):
        return "No description available"

    def copy(self,src,dst,overwrite=False,chunk_size=16384):
        """Copy a file from 'src' to 'dst'.

        src -- The source path
        dst -- The destination path
        overwrite -- If True, then the destination may be overwritten
        (if a file exists at that location). If False then an exception will be
        thrown if the destination exists
        chunk_size -- Size of chunks to use in copy (ignored by S3)
        """
        s3path_dst = self._s3path(dst)
        s3path_dstD = s3path_dst + self._separator
        #  Check for various preconditions.
        ks = self._s3bukt.list(prefix=s3path_dst,delimiter=self._separator)
        dstOK = False
        for k in ks:
            # It exists as a regular file
            if _eq_utf8(k.name,s3path_dst):
                if not overwrite:
                    raise DestinationExistsError(dst)
                dstOK = True
                break
            # Check if it refers to a directory.  If so, we copy *into* it.
            # Since S3 lists in lexicographic order, subsequent iterations
            # of the loop will check for the existence of the new filename.
            if _eq_utf8(k.name,s3path_dstD):
                nm = basename(src)
                dst = pathjoin(dirname(dst),nm)
                s3path_dst = s3path_dstD + nm
                dstOK = True
        if not dstOK and not self.isdir(dirname(dst)):
            msg = "Destination directory does not exist: %(path)s"
            raise ParentDirectoryMissingError(dst,msg=msg)
        # OK, now we can copy the file.
        s3path_src = self._s3path(src)
        try:
            self._s3bukt.copy_key(s3path_dst,self._bucket_name,s3path_src)
        except S3ResponseError, e:
            if "404 Not Found" in str(e):
                msg = "Source is not a file: %(path)s"
                raise ResourceInvalidError(src, msg=msg)
            raise e
        else:
            k = self._s3bukt.get_key(s3path_dst)
            while k is None:
                k = self._s3bukt.get_key(s3path_dst)
            self._sync_key(k)

    def move(self,src,dst,overwrite=False,chunk_size=16384):
        """Move a file from one location to another."""
        self.copy(src,dst,overwrite=overwrite)
        self._s3bukt.delete_key(self._s3path(src))

    def walkfiles(self,
              path="/",
              wildcard=None,
              dir_wildcard=None,
              search="breadth",
              ignore_errors=False ):
        if search != "breadth" or dir_wildcard is not None:
            args = (wildcard,dir_wildcard,search,ignore_errors)
            for item in super(S3FS,self).walkfiles(path,*args):
                yield item
        else:
            prefix = self._s3path(path)
            for k in self._s3bukt.list(prefix=prefix):
                name = relpath(self._uns3path(k.name,prefix))
                if name != "":
                    if not isinstance(name,unicode):
                        name = name.decode("utf8")
                    if not k.name.endswith(self._separator):
                        if wildcard is not None:
                            if callable(wildcard):
                                if not wildcard(basename(name)):
                                    continue
                            else:
                                if not fnmatch(basename(name),wildcard):
                                    continue
                        yield pathjoin(path,name)


    def walkinfo(self,
              path="/",
              wildcard=None,
              dir_wildcard=None,
              search="breadth",
              ignore_errors=False ):
        if search != "breadth" or dir_wildcard is not None:
            args = (wildcard,dir_wildcard,search,ignore_errors)
            for item in super(S3FS,self).walkfiles(path,*args):
                yield (item,self.getinfo(item))
        else:
            prefix = self._s3path(path)
            for k in self._s3bukt.list(prefix=prefix):
                name = relpath(self._uns3path(k.name,prefix))
                if name != "":
                    if not isinstance(name,unicode):
                        name = name.decode("utf8")
                    if wildcard is not None:
                        if callable(wildcard):
                            if not wildcard(basename(name)):
                                continue
                        else:
                            if not fnmatch(basename(name),wildcard):
                                continue
                    yield (pathjoin(path,name),self._get_key_info(k,name))


    def walkfilesinfo(self,
              path="/",
              wildcard=None,
              dir_wildcard=None,
              search="breadth",
              ignore_errors=False ):
        if search != "breadth" or dir_wildcard is not None:
            args = (wildcard,dir_wildcard,search,ignore_errors)
            for item in super(S3FS,self).walkfiles(path,*args):
                yield (item,self.getinfo(item))
        else:
            prefix = self._s3path(path)
            for k in self._s3bukt.list(prefix=prefix):
                name = relpath(self._uns3path(k.name,prefix))
                if name != "":
                    if not isinstance(name,unicode):
                        name = name.decode("utf8")
                    if not k.name.endswith(self._separator):
                        if wildcard is not None:
                            if callable(wildcard):
                                if not wildcard(basename(name)):
                                    continue
                            else:
                                if not fnmatch(basename(name),wildcard):
                                    continue
                        yield (pathjoin(path,name),self._get_key_info(k,name))



def _eq_utf8(name1,name2):
    if isinstance(name1,unicode):
        name1 = name1.encode("utf8")
    if isinstance(name2,unicode):
        name2 = name2.encode("utf8")
    return name1 == name2

def _startswith_utf8(name1,name2):
    if isinstance(name1,unicode):
        name1 = name1.encode("utf8")
    if isinstance(name2,unicode):
        name2 = name2.encode("utf8")
    return name1.startswith(name2)




"""
fs.osfs.watch_inotify
=============

Change watcher support for OSFS, backed by pyinotify.

"""

import os
import sys
import errno
import select
import threading

from fs.errors import *
from fs.path import *
from fs.watch import *

try:
    import pyinotify
except Exception, e:
    #  pyinotify sometimes raises its own custom errors on import.
    #  How on earth are we supposed to catch them when we can't import them?
    if isinstance(e,ImportError):
        raise
    raise ImportError("could not import pyinotify")
try:
    pyinotify.WatchManager.get_fd
except AttributeError:
    raise ImportError("pyinotify version is too old")


class OSFSWatchMixin(WatchableFSMixin):
    """Mixin providing change-watcher support via pyinotify."""

    __watch_lock = threading.Lock()
    __watch_thread = None

    def close(self):
        super(OSFSWatchMixin,self).close()
        self.notify_watchers(CLOSED)
        for watcher_list in self._watchers.values():
            for watcher in watcher_list:
                self.del_watcher(watcher)
        self.__watch_lock.acquire()
        try:
            wt = self.__watch_thread
            if wt is not None and not wt.watchers:
                wt.stop()
                wt.join()
                OSFSWatchMixin.__watch_thread = None
        finally:
            self.__watch_lock.release()

    @convert_os_errors
    def add_watcher(self,callback,path="/",events=None,recursive=True):
        super_add_watcher = super(OSFSWatchMixin,self).add_watcher
        w = super_add_watcher(callback,path,events,recursive)
        w._pyinotify_id = None
        syspath = self.getsyspath(path)
        if isinstance(syspath,unicode):
            syspath = syspath.encode(sys.getfilesystemencoding())
        #  Each watch gets its own WatchManager, since it's tricky to make
        #  a single WatchManager handle multiple callbacks with different
        #  events for a single path.  This means we pay one file descriptor
        #  for each watcher added to the filesystem.  That's not too bad.
        w._pyinotify_WatchManager = wm = pyinotify.WatchManager()
        #  Each individual notifier gets multiplexed by a single shared thread.
        w._pyinotify_Notifier = pyinotify.Notifier(wm)
        evtmask = self.__get_event_mask(events)
        def process_events(event):
            self.__route_event(w,event)
        kwds = dict(rec=recursive,auto_add=recursive,quiet=False)
        try:
            wids = wm.add_watch(syspath,evtmask,process_events,**kwds)
        except pyinotify.WatchManagerError, e:
            raise OperationFailedError("add_watcher",details=e)
        w._pyinotify_id = wids[syspath]
        self.__watch_lock.acquire()
        try:
            wt = self.__get_watch_thread()
            wt.add_watcher(w)
        finally:
            self.__watch_lock.release()
        return w

    @convert_os_errors
    def del_watcher(self,watcher_or_callback):
        if isinstance(watcher_or_callback,Watcher):
            watchers = [watcher_or_callback]
        else:
            watchers = self._find_watchers(watcher_or_callback)
        for watcher in watchers:
            wm = watcher._pyinotify_WatchManager
            wm.rm_watch(watcher._pyinotify_id,rec=watcher.recursive)
            super(OSFSWatchMixin,self).del_watcher(watcher)
        self.__watch_lock.acquire()
        try:
            wt = self.__get_watch_thread()
            for watcher in watchers:
                wt.del_watcher(watcher)
        finally:
            self.__watch_lock.release()

    def __get_event_mask(self,events):
        """Convert the given set of events into a pyinotify event mask."""
        if events is None:
            events = (EVENT,)
        mask = 0
        for evt in events:
            if issubclass(ACCESSED,evt):
                mask |= pyinotify.IN_ACCESS
            if issubclass(CREATED,evt):
                mask |= pyinotify.IN_CREATE
            if issubclass(REMOVED,evt):
                mask |= pyinotify.IN_DELETE
                mask |= pyinotify.IN_DELETE_SELF
            if issubclass(MODIFIED,evt):
                mask |= pyinotify.IN_ATTRIB
                mask |= pyinotify.IN_MODIFY
                mask |= pyinotify.IN_CLOSE_WRITE
            if issubclass(MOVED_SRC,evt):
                mask |= pyinotify.IN_MOVED_FROM
                mask |= pyinotify.IN_MOVED_TO
            if issubclass(MOVED_DST,evt):
                mask |= pyinotify.IN_MOVED_FROM
                mask |= pyinotify.IN_MOVED_TO
            if issubclass(OVERFLOW,evt):
                mask |= pyinotify.IN_Q_OVERFLOW
            if issubclass(CLOSED,evt):
                mask |= pyinotify.IN_UNMOUNT
        return mask

    def __route_event(self,watcher,inevt):
        """Convert pyinotify event into fs.watch event, then handle it."""
        try:
            path = self.unsyspath(inevt.pathname)
        except ValueError:
            return
        try:
            src_path = inevt.src_pathname
            if src_path is not None:
                src_path = self.unsyspath(src_path)
        except (AttributeError,ValueError):
            src_path = None
        if inevt.mask & pyinotify.IN_ACCESS:
            watcher.handle_event(ACCESSED(self,path))
        if inevt.mask & pyinotify.IN_CREATE:
            watcher.handle_event(CREATED(self,path))
            #  Recursive watching of directories in pyinotify requires
            #  the creation of a new watch for each subdir, resulting in
            #  a race condition whereby events in the subdir are missed.
            #  We'd prefer to duplicate events than to miss them.
            if inevt.mask & pyinotify.IN_ISDIR:
                try:
                    #  pyinotify does this for dirs itself, we only.
                    #  need to worry about newly-created files.
                    for child in self.listdir(path,files_only=True):
                        cpath = pathjoin(path,child)
                        self.notify_watchers(CREATED,cpath)
                        self.notify_watchers(MODIFIED,cpath,True)
                except FSError:
                    pass
        if inevt.mask & pyinotify.IN_DELETE:
            watcher.handle_event(REMOVED(self,path))
        if inevt.mask & pyinotify.IN_DELETE_SELF:
            watcher.handle_event(REMOVED(self,path))
        if inevt.mask & pyinotify.IN_ATTRIB:
            watcher.handle_event(MODIFIED(self,path,False))
        if inevt.mask & pyinotify.IN_MODIFY:
            watcher.handle_event(MODIFIED(self,path,True))
        if inevt.mask & pyinotify.IN_CLOSE_WRITE:
            watcher.handle_event(MODIFIED(self,path,True, closed=True))
        if inevt.mask & pyinotify.IN_MOVED_FROM:
            # Sorry folks, I'm not up for decoding the destination path.
            watcher.handle_event(MOVED_SRC(self,path,None))
        if inevt.mask & pyinotify.IN_MOVED_TO:
            if getattr(inevt,"src_pathname",None):
                watcher.handle_event(MOVED_SRC(self,src_path,path))
                watcher.handle_event(MOVED_DST(self,path,src_path))
            else:
                watcher.handle_event(MOVED_DST(self,path,None))
        if inevt.mask & pyinotify.IN_Q_OVERFLOW:
            watcher.handle_event(OVERFLOW(self))
        if inevt.mask & pyinotify.IN_UNMOUNT:
            watcher.handle_event(CLOSED(self))

    def __get_watch_thread(self):
        """Get the shared watch thread, initializing if necessary.

        This method must only be called while holding self.__watch_lock, or
        multiple notifiers could be created.
        """
        if OSFSWatchMixin.__watch_thread is None:
            OSFSWatchMixin.__watch_thread = SharedThreadedNotifier()
            OSFSWatchMixin.__watch_thread.start()
        return OSFSWatchMixin.__watch_thread


class SharedThreadedNotifier(threading.Thread):
    """pyinotifer Notifier that can manage multiple WatchManagers.

    Each watcher added to an OSFS corresponds to a new pyinotify.WatchManager
    instance.  Rather than run a notifier thread for each manager, we run a
    single thread that multiplexes between them all.
    """

    def __init__(self):
        super(SharedThreadedNotifier,self).__init__()
        self.daemon = True
        self.running = True
        self._pipe_r, self._pipe_w = os.pipe()
        self._poller = select.poll()
        self._poller.register(self._pipe_r,select.POLLIN)
        self.watchers = {}

    def add_watcher(self,watcher):
        fd = watcher._pyinotify_WatchManager.get_fd()
        self.watchers[fd] = watcher
        self._poller.register(fd,select.POLLIN)
        #  Bump the poll object so it recognises the new fd.
        os.write(self._pipe_w,b"H")

    def del_watcher(self,watcher):
        fd = watcher._pyinotify_WatchManager.get_fd()
        try:
            del self.watchers[fd]
        except KeyError:
            pass
        else:
            self._poller.unregister(fd)

    def run(self):
        #  Grab some attributes of the select module, so they're available
        #  even when shutting down the interpreter.
        _select_error = select.error
        _select_POLLIN = select.POLLIN
        #  Loop until stopped, dispatching to individual notifiers.
        while self.running:
            try:
                ready_fds = self._poller.poll()
            except _select_error, e:
                if e[0] != errno.EINTR:
                    raise
            else:
                for (fd,event) in ready_fds:
                    #  Ignore all events other than "input ready".
                    if not event & _select_POLLIN:
                        continue
                    #  For signals on our internal pipe, just read and discard.
                    if fd == self._pipe_r:
                        os.read(self._pipe_r,1)
                    #  For notifier fds, dispath to the notifier methods.
                    else:
                        try:
                            notifier = self.watchers[fd]._pyinotify_Notifier
                        except KeyError:
                            pass
                        else:
                            notifier.read_events()
                            try:
                                notifier.process_events()
                            except EnvironmentError:
                                pass

    def stop(self):
        if self.running:
            self.running = False
            os.write(self._pipe_w,"S")
            os.close(self._pipe_w)





#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.api import confirm, error, YES
from traits.api import Instance
from traitsui.api import Group, Handler, Item, SetEditor, View
from traitsui.menu import Action, CancelButton

# Local imports.
from apptools.permissions.package_globals import get_permissions_manager
from apptools.permissions.permission import Permission
from i_policy_storage import PolicyStorageError
from policy_data import Role
from select_role import select_role


class _RoleView(View):
    """The view for handling roles."""

    #### 'View' interface #####################################################

    kind = 'modal'

    title = "Define roles"

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        buttons = [Action(name="Search"), Action(name="Add"),
                Action(name="Modify"), Action(name="Delete"), CancelButton]

        all_perms = get_permissions_manager().policy_manager.permissions.values()

        perms_editor = SetEditor(values=all_perms,
                left_column_title="Available Permissions",
                right_column_title="Assigned Permissions")

        perms_group = Group(Item(name='permissions', editor=perms_editor),
                label='Permissions', show_border=True, show_labels=False)

        super(_RoleView, self).__init__(Item(name='name'),
                Item(name='description'), perms_group, buttons=buttons,
                **traits)


class _RoleHandler(Handler):
    """The view handler for roles."""

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _search_clicked(self, info):
        """Invoked by the "Search" button."""

        role = self._role(info)

        # Get all roles that satisfy the criteria.
        try:
            roles = get_permissions_manager().policy_manager.policy_storage.matching_roles(role.name)
        except PolicyStorageError, e:
            self._ps_error(e)
            return

        if len(roles) == 0:
            self._error("There is no role that matches \"%s\"." % role.name)
            return

        name, description, perm_ids = select_role(roles)

        if name:
            # Update the viewed object.
            role.name = name
            role.description = description
            role.permissions = self._perms_to_list(perm_ids)

    def _add_clicked(self, info):
        """Invoked by the "Add" button."""

        role = self._validate(info)
        if role is None:
            return

        # Add the data to the database.
        try:
            get_permissions_manager().policy_manager.policy_storage.add_role(
                    role.name, role.description,
                    [p.id for p in role.permissions])
            info.ui.dispose()
        except PolicyStorageError, e:
            self._ps_error(e)

    def _modify_clicked(self, info):
        """Invoked by the "Modify" button."""

        role = self._validate(info)
        if role is None:
            return

        # Update the data in the database.
        try:
            get_permissions_manager().policy_manager.policy_storage.modify_role(
                    role.name, role.description,
                    [p.id for p in role.permissions])
            info.ui.dispose()
        except PolicyStorageError, e:
            self._ps_error(e)

    def _delete_clicked(self, info):
        """Invoked by the "Delete" button."""

        role = self._validate(info)
        if role is None:
            return

        if confirm(None, "Are you sure you want to delete the role \"%s\"?" % role.name) == YES:
            # Delete the data from the database.
            try:
                get_permissions_manager().policy_manager.policy_storage.delete_role(role.name)

                info.ui.dispose()
            except PolicyStorageError, e:
                self._ps_error(e)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _validate(self, info):
        """Validate the role and return it if there were no problems."""

        role = self._role(info)
        role.name = role.name.strip()

        if not role.name:
            self._error("A role name must be given.")
            return None

        return role

    def _perms_to_list(self, perm_ids):
        """Return a list of Permission instances created from the given list of
        permission ids."""

        pl = []

        for id in perm_ids:
            try:
                p = get_permissions_manager().policy_manager.permissions[id]
            except KeyError:
                # FIXME: permissions should be populated from the policy
                # database - or is it needed at all?  Should it just be read
                # when managing roles?
                p = Permission(id=id, application_defined=False)

            pl.append(p)

        return pl

    @staticmethod
    def _role(info):
        """Return the role instance being handled."""

        return info.ui.context['object']

    @staticmethod
    def _error(msg):
        """Display an error message to the user."""

        error(None, msg)

    @staticmethod
    def _ps_error(e):
        """Display a message to the user after a PolicyStorageError exception
        has been raised."""

        error(None, str(e))


def role_definition():
    """Implement the role definition for the current policy manager."""

    role = Role()
    view = _RoleView()
    handler = _RoleHandler()

    role.edit_traits(view=view, handler=handler)



"""
fs.zipfs
========

A FS object that represents the contents of a Zip file

"""

import datetime
import os.path

from fs.base import *
from fs.path import *
from fs.errors import *
from fs.filelike import StringIO
from fs import iotools

from zipfile import ZipFile, ZIP_DEFLATED, ZIP_STORED, BadZipfile, LargeZipFile
from memoryfs import MemoryFS

import tempfs

from six import PY3


class ZipOpenError(CreateFailedError):
    """Thrown when the zip file could not be opened"""
    pass


class ZipNotFoundError(CreateFailedError):
    """Thrown when the requested zip file does not exist"""
    pass


class _TempWriteFile(object):
    """Proxies a file object and calls a callback when the file is closed."""

    def __init__(self, fs, filename, close_callback):
        self.fs = fs
        self.filename = filename
        self._file = self.fs.open(filename, 'wb+')
        self.close_callback = close_callback

    def write(self, data):
        return self._file.write(data)

    def tell(self):
        return self._file.tell()

    def close(self):
        self._file.close()
        self.close_callback(self.filename)

    def flush(self):
        self._file.flush()

    def seek(self, offset, whence):
        return self._file.seek(offset, whence)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class _ExceptionProxy(object):
    """A placeholder for an object that may no longer be used."""

    def __getattr__(self, name):
        raise ValueError("Zip file has been closed")

    def __setattr__(self, name, value):
        raise ValueError("Zip file has been closed")

    def __nonzero__(self):
        return False


class ZipFS(FS):
    """A FileSystem that represents a zip file."""

    _meta = {'thread_safe': True,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': False,
             'network': False,
             'atomic.setcontents': False
             }

    def __init__(self, zip_file, mode="r", compression="deflated", allow_zip_64=False, encoding="CP437", thread_synchronize=True):
        """Create a FS that maps on to a zip file.

        :param zip_file: a (system) path, or a file-like object
        :param mode: mode to open zip file, 'r' for reading, 'w' for writing or 'a' for appending
        :param compression: can be 'deflated' (default) to compress data or 'stored' to just store date
        :param allow_zip_64: set to True to use zip files greater than 2 GB, default is False
        :param encoding: the encoding to use for unicode filenames
        :param thread_synchronize: set to True (default) to enable thread-safety
        :raises `fs.errors.ZipOpenError`: thrown if the zip file could not be opened
        :raises `fs.errors.ZipNotFoundError`: thrown if the zip file does not exist (derived from ZipOpenError)

        """
        super(ZipFS, self).__init__(thread_synchronize=thread_synchronize)
        if compression == "deflated":
            compression_type = ZIP_DEFLATED
        elif compression == "stored":
            compression_type = ZIP_STORED
        else:
            raise ValueError("Compression should be 'deflated' (default) or 'stored'")

        if len(mode) > 1 or mode not in "rwa":
            raise ValueError("mode must be 'r', 'w' or 'a'")

        self.zip_mode = mode
        self.encoding = encoding

        if isinstance(zip_file, basestring):
            zip_file = os.path.expanduser(os.path.expandvars(zip_file))
            zip_file = os.path.normpath(os.path.abspath(zip_file))
            self._zip_file_string = True
        else:
            self._zip_file_string = False

        try:
            self.zf = ZipFile(zip_file, mode, compression_type, allow_zip_64)
        except BadZipfile, bzf:
            raise ZipOpenError("Not a zip file or corrupt (%s)" % str(zip_file),
                               details=bzf)
        except IOError, ioe:
            if str(ioe).startswith('[Errno 22] Invalid argument'):
                raise ZipOpenError("Not a zip file or corrupt (%s)" % str(zip_file),
                                   details=ioe)
            raise ZipNotFoundError("Zip file not found (%s)" % str(zip_file),
                                   details=ioe)

        self.zip_path = str(zip_file)
        self.temp_fs = None
        if mode in 'wa':
            self.temp_fs = tempfs.TempFS()

        self._path_fs = MemoryFS()
        if mode in 'ra':
            self._parse_resource_list()

        self.read_only = mode == 'r'

    def __str__(self):
        return "<ZipFS: %s>" % self.zip_path

    def __unicode__(self):
        return u"<ZipFS: %s>" % self.zip_path

    def _decode_path(self, path):
        if PY3:
            return path
        return path.decode(self.encoding)

    def _encode_path(self, path):
        if PY3:
            return path
        return path.encode(self.encoding)

    def _parse_resource_list(self):
        for path in self.zf.namelist():
            #self._add_resource(path.decode(self.encoding))
            self._add_resource(self._decode_path(path))

    def _add_resource(self, path):
        if path.endswith('/'):
            path = path[:-1]
            if path:
                self._path_fs.makedir(path, recursive=True, allow_recreate=True)
        else:
            dirpath, _filename = pathsplit(path)
            if dirpath:
                self._path_fs.makedir(dirpath, recursive=True, allow_recreate=True)
            f = self._path_fs.open(path, 'w')
            f.close()

    def getmeta(self, meta_name, default=NoDefaultMeta):
        if meta_name == 'read_only':
            return self.read_only
        return super(ZipFS, self).getmeta(meta_name, default)

    def close(self):
        """Finalizes the zip file so that it can be read.
        No further operations will work after this method is called."""

        if hasattr(self, 'zf') and self.zf:
            self.zf.close()
            self.zf = _ExceptionProxy()

    @synchronize
    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        path = normpath(relpath(path))

        if 'r' in mode:
            if self.zip_mode not in 'ra':
                raise OperationFailedError("open file",
                                           path=path,
                                           msg="1 Zip file must be opened for reading ('r') or appending ('a')")
            try:
                if hasattr(self.zf, 'open') and self._zip_file_string:
                    #return self.zf.open(self._encode_path(path), "r")
                    return self.zf.open(self._encode_path(path), 'rU' if 'U' in mode else 'r')
                else:
                    contents = self.zf.read(self._encode_path(path))
            except KeyError:
                raise ResourceNotFoundError(path)
            return StringIO(contents)

        if 'w' in mode:
            if self.zip_mode not in 'wa':
                raise OperationFailedError("open file",
                                           path=path,
                                           msg="2 Zip file must be opened for writing ('w') or appending ('a')")
            dirname, _filename = pathsplit(path)
            if dirname:
                self.temp_fs.makedir(dirname, recursive=True, allow_recreate=True)

            self._add_resource(path)
            f = _TempWriteFile(self.temp_fs, path, self._on_write_close)
            return f

        raise ValueError("Mode must contain be 'r' or 'w'")

    @synchronize
    def getcontents(self, path, mode="rb", encoding=None, errors=None, newline=None):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = normpath(relpath(path))
        try:
            contents = self.zf.read(self._encode_path(path))
        except KeyError:
            raise ResourceNotFoundError(path)
        except RuntimeError:
            raise OperationFailedError("read file", path=path, msg="3 Zip file must be opened with 'r' or 'a' to read")
        if 'b' in mode:
            return contents
        return iotools.decode_binary(contents, encoding=encoding, errors=errors, newline=newline)

    @synchronize
    def _on_write_close(self, filename):
        sys_path = self.temp_fs.getsyspath(filename)
        self.zf.write(sys_path, self._encode_path(filename))

    def desc(self, path):
        return "%s in zip file %s" % (path, self.zip_path)

    def isdir(self, path):
        return self._path_fs.isdir(path)

    def isfile(self, path):
        return self._path_fs.isfile(path)

    def exists(self, path):
        return self._path_fs.exists(path)

    @synchronize
    def makedir(self, dirname, recursive=False, allow_recreate=False):
        dirname = normpath(dirname)
        if self.zip_mode not in "wa":
            raise OperationFailedError("create directory", path=dirname, msg="4 Zip file must be opened for writing ('w') or appending ('a')")
        if not dirname.endswith('/'):
            dirname += '/'
        self._add_resource(dirname)

    def listdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        return self._path_fs.listdir(path, wildcard, full, absolute, dirs_only, files_only)

    @synchronize
    def getinfo(self, path):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = normpath(path).lstrip('/')
        try:
            zi = self.zf.getinfo(self._encode_path(path))
            zinfo = dict((attrib, getattr(zi, attrib)) for attrib in dir(zi) if not attrib.startswith('_'))
            for k, v in zinfo.iteritems():
                if callable(v):
                    zinfo[k] = v()
        except KeyError:
            zinfo = {'file_size': 0}
        info = {'size': zinfo['file_size']}
        if 'date_time' in zinfo:
            info['modified_time'] = info['created_time'] = datetime.datetime(*zinfo['date_time'])
        info.update(zinfo)
        if 'FileHeader' in info:
            del info['FileHeader']
        return info



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.etsconfig.api import ETSConfig
from traitsui.api import Handler

# Local imports.
from adapter_base import AdapterBase
from package_globals import get_permissions_manager


# Register the bundled adapters.
from adapters import pyface_action

if ETSConfig.toolkit == 'wx':
    from adapters import wx_window
elif ETSConfig.toolkit == 'qt4':
    from adapters import qt4_widget


class SecureProxy(object):
    """The SecureProxy class is a wrapper for an object whose enabled and
    visible states can be managed.  It attaches one or more permissions to the
    object and enables and shows the object only if those permissions allow it.
    In all other respects it behaves exactly like the object.  It is based on
    Tomer Filiba's Object Proxy cookbook recipe."""

    __slots__ = ('_ets', '__weakref__')

    def __init__(self, proxied, permissions, show=True):
        """Initialise the instance.  proxied is the object whose enabled and
        visible states are managed according to the permissions of the current
        user.  permissions is a list of permissions to attach to the object.
        show is set if the proxied object should be visible when it is
        disabled."""

        adapter = object.__getattribute__(self, '_ets')

        # Correct the current values.
        if not show:
            adapter.update_visible(adapter.get_visible())

        adapter.update_enabled(adapter.get_enabled())

    # Proxying (special cases).
    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, '_ets').proxied, name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, '_ets').proxied, name)

    def __setattr__(self, name, value):
        object.__getattribute__(self, '_ets').setattr(name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, '_ets').proxied)

    def __str__(self):
        return str(object.__getattribute__(self, '_ets').proxied)

    def __repr__(self):
        return repr(object.__getattribute__(self, '_ets').proxied)

    # Factories.
    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__',
        '__irshift__', '__isub__', '__iter__', '__itruediv__', '__ixor__',
        '__le__', '__len__', '__long__', '__lshift__', '__lt__', '__mod__',
        '__mul__', '__ne__', '__neg__', '__oct__', '__or__', '__pos__',
        '__pow__', '__radd__', '__rand__', '__rdiv__', '__rdivmod__',
        '__reduce__', '__reduce_ex__', '__repr__', '__reversed__',
        '__rfloorfiv__', '__rlshift__', '__rmod__', '__rmul__', '__ror__',
        '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__',
        '__rxor__', '__setitem__', '__setslice__', '__sub__', '__truediv__',
        '__xor__', 'next',
    ]

    @classmethod
    def _ets_class_proxy(cls, theclass):
        """Creates a proxy for the given class."""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, '_ets').proxied, name)(*args, **kw)
            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name) and not hasattr(cls, name):
                namespace[name] = make_method(name)

        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,),
                namespace)

    def __new__(cls, proxied, permissions, show=True):
        """Apply a set of permissions to an object.  This may be done by
        creating a proxy or by modifying the object in situ depending on its
        type.
        """

        # Create the adapter.
        adapter = AdapterBase.factory(proxied, permissions, show)

        # Try and adapt the object itself.
        adapted = adapter.adapt()

        if adapted is None:
            # Create a wrapper for the object.  The cache is unique per
            # deriving class to ensure there are no clashes with any
            # sub-classes with the same name.
            try:
                cache = cls.__dict__['_ets_cache']
            except KeyError:
                cls._ets_cache = cache = {}

            pclass = proxied.__class__
            try:
                theclass = cache[pclass]
            except KeyError:
                cache[pclass] = theclass = cls._ets_class_proxy(pclass)

            adapted = object.__new__(theclass)

            # Save the adapter in the adapted object.
            object.__setattr__(adapted, '_ets', adapter)
        else:
            # Correct the current values.
            if not show:
                adapter.update_visible(adapter.get_visible())

            adapter.update_enabled(adapter.get_enabled())

            # Save the adapter in the adapted object.
            adapted._ets = adapter

        return adapted


class SecureHandler(Handler):
    """The SecureHandler class is a sub-class of the TraitsUI Handler class
    that ensures that the enabled and visible state of the items of a TraitsUI
    view are updated when the user's authorisation state changes.
    """

    def __init__(self, **traits):
        """Initialise the object."""

        super(SecureHandler, self).__init__(**traits)

        get_permissions_manager().user_manager.on_trait_event(self._refresh,
                'user_authenticated')

    def init_info(self, info):
        """Reimplemented to save the UIInfo object."""

        self._info = info

    def _refresh(self):
        """Invoked whenever the current user's authorisation state changes."""

        # FIXME: This is (currently) an internal method.
        self._info.ui._evaluate_when()



""" Skeleton sample task

"""
# Enthought library imports.
from pyface.api import ImageResource
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import DockPaneToggleGroup, SMenuBar, \
    SMenu, SToolBar, TaskAction, TaskToggleGroup
from traits.api import on_trait_change, Property, Instance

from panes import SkeletonPane1, SkeletonPane2, SkeletonPane3


class SkeletonTask(Task):
    """ A simple task for opening a blank editor.
    """

    #### Task interface #######################################################

    id = 'omnivore.framework.text_edit_task'
    name = 'Skeleton'

    active_editor = Property(Instance(IEditor),
                             depends_on='editor_area.active_editor')

    editor_area = Instance(IEditorAreaPane)

    menu_bar = SMenuBar(SMenu(TaskAction(name='New', method='new',
                                         accelerator='Ctrl+N'),
                              id='File', name='&File'),
                        SMenu(id='Edit', name='&Edit'),
                        SMenu(#DockPaneToggleGroup(),
                              TaskToggleGroup(),
                              id='View', name='&View'))

    tool_bars = [ SToolBar(TaskAction(method='new',
                                      tooltip='New file',
                                      image=ImageResource('document_new')),
                           image_size = (32, 32)), ]

    ###########################################################################
    # 'Task' interface.
    ###########################################################################

    def _default_layout_default(self):
        return TaskLayout(
            left=VSplitter(
                PaneItem('text_edit.pane1'),
                HSplitter(
                    PaneItem('text_edit.pane2'),
                    PaneItem('text_edit.pane3'),
                    ),
                ))

    def create_central_pane(self):
        """ Create the central pane: the text editor.
        """
        self.editor_area = EditorAreaPane()
        return self.editor_area

    def create_dock_panes(self):
        """ Create the file browser and connect to its double click event.
        """
        return [ SkeletonPane1(), SkeletonPane2(), SkeletonPane3() ]

    ###########################################################################
    # 'ExampleTask' interface.
    ###########################################################################

    def new(self):
        """ Opens a new empty window
        """
        editor = Editor()
        self.editor_area.add_editor(editor)
        self.editor_area.activate_editor(editor)
        self.activated()

    #### Trait property getter/setters ########################################

    def _get_active_editor(self):
        if self.editor_area is not None:
            return self.editor_area.active_editor
        return None



from __future__ import absolute_import

from traits.testing.unittest_tools import UnittestTools, unittest

from ..action import Action
from ..action_item import ActionItem
from ..action_manager import ActionManager
from ..group import Group


class TestActionItem(unittest.TestCase, UnittestTools):

    def setUp(self):
        # test whether function is called by updating list
        # XXX should really use mock
        self.memo = []

        def perform():
            self.memo.append('called')

        self.perform = perform

        self.action = Action(name='Test', on_perform=perform)
        self.action_item = ActionItem(action=self.action)
        self.group = Group(id='test')

    def test_init_group(self):
        action_manager = ActionManager(self.group)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [self.group, default_group])
        self.assertEqual(self.group.parent, action_manager)

    def test_init_string(self):
        action_manager = ActionManager("Test")
        default_group = action_manager._get_default_group()
        self.assertEqual(len(action_manager.groups), 2)
        self.assertEqual(action_manager.groups[0].id, "Test")
        self.assertEqual(action_manager.groups[-1], default_group)
        self.assertEqual(action_manager.groups[0].parent, action_manager)

    def test_init_action_item(self):
        action_manager = ActionManager(self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group])
        self.assertEqual(default_group.items, [self.action_item])
        self.assertEqual(self.action_item.parent, default_group)

    def test_init_group_action_item(self):
        action_manager = ActionManager(self.group, self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [self.group, default_group])
        self.assertEqual(self.group.items, [self.action_item])
        self.assertEqual(self.action_item.parent, self.group)

    def test_init_action(self):
        action_manager = ActionManager(self.action)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group])
        self.assertEqual(default_group.items[0].action, self.action)
        self.assertEqual(default_group.items[0].parent, default_group)

    def test_init_nothing(self):
        action_manager = ActionManager()
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group])
        self.assertEqual(len(default_group.items), 0)

    def test_append(self):
        action_manager = ActionManager()
        action_manager.append(self.group)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group, self.group])
        self.assertEqual(self.group.parent, action_manager)

    def test_append_2(self):
        action_manager = ActionManager(self.group)
        group2 = Group()
        action_manager.append(group2)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [self.group, default_group, group2])
        self.assertEqual(group2.parent, action_manager)

    def test_append_string(self):
        action_manager = ActionManager(self.group)
        action_manager.append("Test string")
        self.assertEqual(len(action_manager.groups), 3)
        self.assertEqual(action_manager.groups[2].id, "Test string")

    def test_append_item(self):
        action_manager = ActionManager()
        action_manager.append(self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group])
        self.assertEqual(default_group.items, [self.action_item])

    def test_append_item_2(self):
        action_manager = ActionManager()
        action_manager.append(self.group)
        action_manager.append(self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group, self.group])
        self.assertEqual(default_group.items, [self.action_item])

    def test_destroy(self):
        action_manager = ActionManager(self.group)
        # XXX items doesn't fire a change event.  Should it?
        # XXX should mock group to ensure that destroy is called
        action_manager.destroy()

    def test_insert(self):
        action_manager = ActionManager()
        action_manager.insert(0, self.group)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [self.group, default_group])
        self.assertEqual(self.group.parent, action_manager)

    def test_insert_2(self):
        action_manager = ActionManager(self.group)
        group2 = Group()
        action_manager.insert(1, group2)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [self.group, group2, default_group])
        self.assertEqual(group2.parent, action_manager)

    def test_insert_string(self):
        action_manager = ActionManager(self.group)
        action_manager.insert(0, "Test string")
        self.assertEqual(len(action_manager.groups), 3)
        self.assertEqual(action_manager.groups[0].id, "Test string")

    def test_insert_item(self):
        action_manager = ActionManager()
        action_manager.insert(0, self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group])
        self.assertEqual(default_group.items, [self.action_item])

    def test_insert_item_2(self):
        action_manager = ActionManager()
        action_manager.append(self.group)
        action_manager.insert(0, self.action_item)
        default_group = action_manager._get_default_group()
        self.assertEqual(action_manager.groups, [default_group, self.group])
        self.assertEqual(default_group.items, [self.action_item])

    def test_find_group(self):
        action_manager = ActionManager(self.group)
        group = action_manager.find_group("test")
        self.assertEqual(group, self.group)

    def test_find_group_missing(self):
        action_manager = ActionManager(self.group)
        group = action_manager.find_group("not here")
        self.assertIsNone(group)

    def test_find_item(self):
        self.group.append(self.action_item)
        action_manager = ActionManager(self.group)
        item = action_manager.find_item("Test")
        self.assertEqual(item, self.action_item)

    def test_find_item_missing(self):
        self.group.append(self.action_item)
        action_manager = ActionManager(self.group)
        item = action_manager.find_item("Not here")
        self.assertIsNone(item)

    def test_find_item_hierarchy(self):
        action_manager = ActionManager(self.group)
        action_manager_2 = ActionManager(self.action_item, id='test2')
        self.group.append(action_manager_2)
        item = action_manager.find_item("test2/Test")
        self.assertEqual(item, self.action_item)

    def test_walk_hierarchy(self):
        action_manager = ActionManager(self.group)
        action_manager_2 = ActionManager(self.action_item, id='test2')
        self.group.append(action_manager_2)
        result = []
        action_manager.walk(result.append)
        self.assertEqual(result,
                         [action_manager,
                            self.group,
                                action_manager_2,
                                    action_manager_2._get_default_group(),
                                        self.action_item,
                            action_manager._get_default_group()])

    def test_enabled_changed(self):
        self.group.append(self.action_item)
        action_manager = ActionManager(self.group)
        action_manager.enabled = False
        self.assertFalse(self.group.enabled)
        self.assertFalse(self.action_item.enabled)
        self.assertFalse(self.action.enabled)
        action_manager.enabled = True
        self.assertTrue(self.group.enabled)
        self.assertTrue(self.action_item.enabled)
        self.assertTrue(self.action.enabled)

    def test_visible_changed(self):
        self.group.append(self.action_item)
        action_manager = ActionManager(self.group)
        action_manager.visible = False
        self.assertFalse(self.group.visible)
        # XXX group doesn't make items invisible
        # self.assertFalse(self.action_item.enabled)
        # self.assertFalse(self.action.enabled)
        action_manager.visible = True
        self.assertTrue(self.group.visible)
        # XXX group doesn't make items visible
        #self.assertTrue(self.action_item.enabled)
        #self.assertTrue(self.action.enabled)



""" A type system with standard(ish) Python semantics. """


# Standard library imports.
import inspect

# Local imports.
from abstract_type_system import AbstractTypeSystem


class PythonObject:
    """ The root type in the type system.

    fixme: Python is currently a bit broken as it has dual type hierarchies,
    one for old-style and one for new-style classes. This class is used to
    create a fake root to unify them.

    """

    __class__ = type


class PythonTypeSystem(AbstractTypeSystem):
    """ A type system with standard(ish) Python semantics. """

    ###########################################################################
    # 'AbstractTypeSystem' interface.
    ###########################################################################

    def is_a(self, obj, type):
        """ Is an object and instance of the specified type? """

        return isinstance(obj, type) or type is PythonObject

    def get_mro(self, type):
        """ Returns the MRO of a type. """

        return list(inspect.getmro(type)) + [PythonObject]

#### EOF ######################################################################



#  Test the 'trait_set', 'trait_get' interface to
#  the HasTraits class.
#
#  Copyright (c) 2014, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  License included in /LICENSE.txt and may be redistributed only under the
#  conditions described in the aforementioned license.  The license is also
#  available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
from __future__ import absolute_import

from contextlib import contextmanager
import logging

import traits.has_traits
from traits.testing.unittest_tools import unittest, UnittestTools

from ..api import HasTraits, Str, Int


class TraitsObject(HasTraits):

    string = Str
    integer = Int


class ListHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.records = []

    def emit(self, record):
        self.records.append(record)


@contextmanager
def catch_logs(module_name):
    logger = logging.getLogger(module_name)
    logger.propagate = False
    handler = ListHandler()
    logger.addHandler(handler)
    try:
        yield handler
    finally:
        logger.removeHandler(handler)
        logger.propagate = True


class TestTraitGet(UnittestTools, unittest.TestCase):

    def test_trait_set(self):
        obj = TraitsObject()
        obj.trait_set(string='foo')
        self.assertEqual(obj.string, 'foo')
        self.assertEqual(obj.integer, 0)

    def test_trait_get(self):
        obj = TraitsObject()
        obj.trait_set(string='foo')
        values = obj.trait_get('string', 'integer')
        self.assertEqual(values, {'string': 'foo', 'integer': 0})

    def test_trait_set_deprecated(self):
        obj = TraitsObject()

        with self.assertNotDeprecated():
            obj.trait_set(integer=1)

        with self.assertDeprecated():
            obj.set(string='foo')

        self.assertEqual(obj.string, 'foo')
        self.assertEqual(obj.integer, 1)

    def test_trait_get_deprecated(self):
        obj = TraitsObject()
        obj.string = 'foo'
        obj.integer = 1

        with self.assertNotDeprecated():
            values = obj.trait_get('integer')
        self.assertEqual(values, {'integer': 1})

        with self.assertDeprecated():
            values = obj.get('string')
        self.assertEqual(values, {'string': 'foo'})



#------------------------------------------------------------------------------
#
#  Copyright (c) 2006, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   01/27/2006
#
#------------------------------------------------------------------------------

""" Defines a source code editor for the wxPython user interface toolkit,
    useful for tools such as debuggers.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx
import wx.stc as stc

from traits.api \
    import Str, List, Int, Event, Bool, TraitError, on_trait_change

from traits.trait_base \
    import SequenceTypes

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.code_editor file.
from traitsui.editors.code_editor \
    import ToolkitEditorFactory

from pyface.api \
    import PythonEditor

from pyface.util.python_stc \
    import faces

from editor \
    import Editor

from constants \
    import OKColor, ErrorColor

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Marker line constants:

# Marks a marked line
MARK_MARKER = 0

# Marks a line matching the current search
SEARCH_MARKER = 1

# Marks the currently selected line
SELECTED_MARKER = 2

#-------------------------------------------------------------------------
#  'SourceEditor' class:
#-------------------------------------------------------------------------


class SourceEditor(Editor):
    """ Editor for source code, which displays a PyFace PythonEditor.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The code editor is scrollable. This value overrides the default.
    scrollable = True

    # Is the editor read only?
    readonly = Bool(False)

    # The currently selected line
    selected_line = Int

    # The currently selected text
    selected_text = Str

    # The list of line numbers to mark
    mark_lines = List(Int)

    # The current line number
    line = Event

    # The current column
    column = Event

    # calltip clicked event
    calltip_clicked = Event

    # The STC lexer use
    lexer = Int

    # The lines to be dimmed
    dim_lines = List(Int)
    dim_color = Str
    _dim_style_number = Int(16)  # 0-15 are reserved for the python lexer

    # The lines to have squiggles drawn under them
    squiggle_lines = List(Int)
    squiggle_color = Str

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        self._editor = editor = PythonEditor(
            parent, show_line_numbers=factory.show_line_numbers)
        self.control = control = editor.control

        # There are a number of events which aren't well documented that look
        # to be useful in future implmentations, below are a subset of the
        # events that look interesting:
        #    EVT_STC_AUTOCOMP_SELECTION
        #    EVT_STC_HOTSPOT_CLICK
        #    EVT_STC_HOTSPOT_DCLICK
        #    EVT_STC_DOUBLECLICK
        #    EVT_STC_MARGINCLICK

        control.SetSize(wx.Size(300, 124))

        # Clear out the goofy hotkeys for zooming text
        control.CmdKeyClear(ord('B'), stc.STC_SCMOD_CTRL)
        control.CmdKeyClear(ord('N'), stc.STC_SCMOD_CTRL)

        # Set up the events
        wx.EVT_KILL_FOCUS(control, self.wx_update_object)
        stc.EVT_STC_CALLTIP_CLICK(control, control.GetId(),
                                  self._calltip_clicked)

        if factory.auto_scroll and (factory.selected_line != ''):
            wx.EVT_SIZE(control, self._update_selected_line)

        if factory.auto_set:
            editor.on_trait_change(self.update_object, 'changed',
                                   dispatch='ui')

        if factory.key_bindings is not None:
            editor.on_trait_change(self.key_pressed, 'key_pressed',
                                   dispatch='ui')

        if self.readonly:
            control.SetReadOnly(True)

        # Set up the lexer
        control.SetLexer(stc.STC_LEX_CONTAINER)
        control.Bind(stc.EVT_STC_STYLENEEDED, self._style_needed)
        try:
            self.lexer = getattr(stc, 'STC_LEX_' + self.factory.lexer.upper())
        except AttributeError:
            self.lexer = stc.STC_LEX_NULL

        # Define the markers we use:
        control.MarkerDefine(MARK_MARKER, stc.STC_MARK_BACKGROUND,
                             background=factory.mark_color_)
        control.MarkerDefine(SEARCH_MARKER, stc.STC_MARK_BACKGROUND,
                             background=factory.search_color_)
        control.MarkerDefine(SELECTED_MARKER, stc.STC_MARK_BACKGROUND,
                             background=factory.selected_color_)

        # Make sure the editor has been initialized:
        self.update_editor()

        # Set up any event listeners:
        self.sync_value(factory.mark_lines, 'mark_lines', 'from',
                        is_list=True)
        self.sync_value(factory.selected_line, 'selected_line', 'from')
        self.sync_value(factory.selected_text, 'selected_text', 'to')
        self.sync_value(factory.line, 'line')
        self.sync_value(factory.column, 'column')
        self.sync_value(factory.calltip_clicked, 'calltip_clicked')

        self.sync_value(factory.dim_lines, 'dim_lines', 'from', is_list=True)
        if self.factory.dim_color == '':
            self.dim_color = 'dark grey'
        else:
            self.sync_value(factory.dim_color, 'dim_color', 'from')

        self.sync_value(factory.squiggle_lines, 'squiggle_lines', 'from',
                        is_list=True)
        if factory.squiggle_color == '':
            self.squiggle_color = 'red'
        else:
            self.sync_value(factory.squiggle_color, 'squiggle_color', 'from')

        # Check if we need to monitor the line or column position being
        # changed:
        if (factory.line != '') or (factory.column != '') or \
                (factory.selected_text != ''):
            stc.EVT_STC_UPDATEUI(control, control.GetId(),
                                 self._position_changed)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Handles the user entering input data in the edit control:
    #-------------------------------------------------------------------------

    def wx_update_object(self, event):
        """ Handles the user entering input data in the edit control.
        """
        self.update_object()
        event.Skip()

    def update_object(self):
        """ Handles the user entering input data in the edit control.
        """
        if not self._locked:
            try:
                value = self.control.GetText()
                if isinstance(self.value, SequenceTypes):
                    value = value.split()
                self.value = value
                self.control.SetBackgroundColour(OKColor)
                self.control.Refresh()
            except TraitError as excp:
                pass

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self._locked = True
        new_value = self.value
        if isinstance(new_value, SequenceTypes):
            new_value = '\n'.join([line.rstrip() for line in new_value])
        control = self.control
        if control.GetText() != new_value:
            readonly = control.GetReadOnly()
            control.SetReadOnly(False)
            l1 = control.GetFirstVisibleLine()
            pos = control.GetCurrentPos()
            control.SetText(new_value)
            control.GotoPos(pos)
            control.ScrollToLine(l1)
            control.SetReadOnly(readonly)
            self._mark_lines_changed()
            self._selected_line_changed()
            self._style_document()

        self._locked = False

    #-------------------------------------------------------------------------
    #  Handles the calltip being clicked:
    #-------------------------------------------------------------------------

    def _calltip_clicked(self, event):
        self.calltip_clicked = True

    #-------------------------------------------------------------------------
    #  Handles the set of 'marked lines' being changed:
    #-------------------------------------------------------------------------

    def _mark_lines_changed(self):
        """ Handles the set of marked lines being changed.
        """
        lines = self.mark_lines
        control = self.control
        lc = control.GetLineCount()
        control.MarkerDeleteAll(MARK_MARKER)

        for line in lines:
            if 0 < line <= lc:
                control.MarkerAdd(line - 1, MARK_MARKER)

        control.Refresh()

    def _mark_lines_items_changed(self):
        self._mark_lines_changed()

    #-------------------------------------------------------------------------
    #  Handles the currently 'selected line' being changed:
    #-------------------------------------------------------------------------

    def _selected_line_changed(self):
        """ Handles a change in which line is currently selected.
        """
        line = self.selected_line
        control = self.control
        line = max(1, min(control.GetLineCount(), line)) - 1
        control.MarkerDeleteAll(SELECTED_MARKER)
        control.MarkerAdd(line, SELECTED_MARKER)
        control.GotoLine(line)
        if self.factory.auto_scroll:
            control.ScrollToLine(line - (control.LinesOnScreen() / 2))

        control.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the 'line' trait being changed:
    #-------------------------------------------------------------------------

    def _line_changed(self, line):
        if not self._locked:
            self.control.GotoLine(line - 1)

    #-------------------------------------------------------------------------
    #  Handles the 'column' trait being changed:
    #-------------------------------------------------------------------------

    def _column_changed(self, column):
        if not self._locked:
            control = self.control
            line = control.LineFromPosition(control.GetCurrentPos())
            control.GotoPos(control.PositionFromLine(line) + column - 1)

    #-------------------------------------------------------------------------
    #  Handles the cursor position being changed:
    #-------------------------------------------------------------------------

    def _position_changed(self, event):
        """ Handles the cursor position being changed.
        """
        control = self.control
        pos = control.GetCurrentPos()
        line = control.LineFromPosition(pos)
        self._locked = True
        self.line = line + 1
        self.column = pos - control.PositionFromLine(line) + 1
        self._locked = False
        self.selected_text = control.GetSelectedText()

    #-------------------------------------------------------------------------
    #  Handles a key being pressed within the editor:
    #-------------------------------------------------------------------------

    def key_pressed(self, event):
        """ Handles a key being pressed within the editor.
        """
        self.factory.key_bindings.do(event.event, self.ui.handler,
                                     self.ui.info)

    #-------------------------------------------------------------------------
    #  Handles the styling of the editor:
    #-------------------------------------------------------------------------

    def _dim_color_changed(self):
        self.control.StyleSetForeground(self._dim_style_number, self.dim_color)
        self.control.StyleSetFaceName(self._dim_style_number, "courier new")
        self.control.StyleSetSize(self._dim_style_number, faces['size'])
        self.control.Refresh()

    def _squiggle_color_changed(self):
        self.control.IndicatorSetStyle(2, stc.STC_INDIC_SQUIGGLE)
        self.control.IndicatorSetForeground(2, self.squiggle_color)
        self.control.Refresh()

    @on_trait_change('dim_lines, squiggle_lines')
    def _style_document(self):
        """ Force the STC to fire a STC_STYLENEEDED event for the entire
            document.
        """
        self.control.ClearDocumentStyle()
        self.control.Colourise(0, -1)
        self.control.Refresh()

    def _style_needed(self, event):
        """ Handles an STC request for styling for some area.
        """
        position = self.control.GetEndStyled()
        start_line = self.control.LineFromPosition(position)
        end = event.GetPosition()
        end_line = self.control.LineFromPosition(end)

        # Fixes a strange a bug with the STC widget where creating a new line
        # after a dimmed line causes it to mysteriously lose its styling
        if start_line in self.dim_lines:
            start_line -= 1

        # Trying to Colourise only the lines that we want does not seem to work
        # so we do the whole area and then override the styling on certain
        # lines
        if self.lexer != stc.STC_LEX_NULL:
            self.control.SetLexer(self.lexer)
            self.control.Colourise(position, end)
            self.control.SetLexer(stc.STC_LEX_CONTAINER)

        for line in xrange(start_line, end_line + 1):
            # We don't use LineLength here because it includes newline
            # characters. Styling these leads to strange behavior.
            position = self.control.PositionFromLine(line)
            style_length = self.control.GetLineEndPosition(line) - position

            if line + 1 in self.dim_lines:
                # Set styling mask to only style text bits, not indicator bits
                self.control.StartStyling(position, 0x1f)
                self.control.SetStyling(style_length, self._dim_style_number)
            elif self.lexer == stc.STC_LEX_NULL:
                self.control.StartStyling(position, 0x1f)
                self.control.SetStyling(style_length, stc.STC_STYLE_DEFAULT)

            if line + 1 in self.squiggle_lines:
                self.control.StartStyling(position, stc.STC_INDIC2_MASK)
                self.control.SetStyling(style_length, stc.STC_INDIC2_MASK)
            else:
                self.control.StartStyling(position, stc.STC_INDIC2_MASK)
                self.control.SetStyling(style_length, stc.STC_STYLE_DEFAULT)

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self.control.SetBackgroundColour(ErrorColor)
        self.control.Refresh()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self.factory.auto_set:
            self._editor.on_trait_change(self.update_object, 'changed',
                                         remove=True)
        if self.factory.key_bindings is not None:
            self._editor.on_trait_change(self.key_pressed, 'key_pressed',
                                         remove=True)

        wx.EVT_KILL_FOCUS(self.control, None)

        super(SourceEditor, self).dispose()

    #-- UI preference save/restore interface ---------------------------------

    #-------------------------------------------------------------------------
    #  Restores any saved user preference information associated with the
    #  editor:
    #-------------------------------------------------------------------------

    def restore_prefs(self, prefs):
        """ Restores any saved user preference information associated with the
            editor.
        """
        if self.factory.key_bindings is not None:
            key_bindings = prefs.get('key_bindings')
            if key_bindings is not None:
                self.factory.key_bindings.merge(key_bindings)

    #-------------------------------------------------------------------------
    #  Returns any user preference information associated with the editor:
    #-------------------------------------------------------------------------

    def save_prefs(self):
        """ Returns any user preference information associated with the editor.
        """
        return {'key_bindings': self.factory.key_bindings}

# Define the simple, custom, text and readonly editors, which will be accessed
# by the editor factory for code editors.

CustomEditor = SimpleEditor = TextEditor = SourceEditor


class ReadonlyEditor(SourceEditor):

    # Set the value of the readonly trait.
    readonly = True

### EOF ##################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/22/2004
#
#------------------------------------------------------------------------------

""" Defines constants used by the wxPython implementation of the various text
    editors and text editor factories.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import sys

import wx

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Define platform and wx version constants:
is_mac = (sys.platform == 'darwin')
is_wx26 = (float('.'.join(wx.__version__.split('.')[0:2])) < 2.8)

# Default dialog title
DefaultTitle = 'Edit properties'

# Color of valid input
OKColor = wx.WHITE

# Color to highlight input errors
ErrorColor = wx.Colour(255, 192, 192)

# Color for background of read-only fields
ReadonlyColor = wx.Colour(244, 243, 238)

# Color for background of fields where objects can be dropped
DropColor = wx.Colour(215, 242, 255)

# Color for an editable field
EditableColor = wx.WHITE

# Color for background of windows (like dialog background color)
if is_mac:
    WindowColor = wx.Colour(232, 232, 232)
    BorderedGroupColor = wx.Colour(224, 224, 224)
else:
    WindowColor = wx.SystemSettings_GetColour(wx.SYS_COLOUR_MENUBAR)

# Standard width of an image bitmap
standard_bitmap_width = 120

# Width of a scrollbar
scrollbar_dx = wx.SystemSettings_GetMetric(wx.SYS_VSCROLL_X)

# Screen size values:
screen_dx = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
screen_dy = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------

from __future__ import absolute_import

# Enthought library imports.
from traits.api import Bool, Event, HasTraits, Instance, Int, Property, \
    Unicode, provides

# Local imports.
from .i_undo_manager import IUndoManager


@provides(IUndoManager)
class UndoManager(HasTraits):
    """ The UndoManager class is the default implementation of the
    IUndoManager interface.
    """

    #### 'IUndoManager' interface #############################################

    # This is the currently active command stack and may be None.  Typically it
    # is set when some sort of editor becomes active.
    active_stack = Instance('apptools.undo.api.ICommandStack')

    # This reflects the clean state of the currently active command stack.  It
    # is intended to support a "document modified" indicator in the GUI.  It is
    # maintained by the undo manager.
    active_stack_clean = Property(Bool)

    # This is the name of the command that can be redone.  It will be empty if
    # there is no command that can be redone.  It is maintained by the undo
    # manager.
    redo_name = Property(Unicode)

    # This is the sequence number of the next command to be performed.  It is
    # incremented immediately before a command is invoked (by its 'do()'
    # method).
    sequence_nr = Int

    # This event is fired when the index of a command stack changes.  The value
    # of the event is the stack that has changed.  Note that it may not be the
    # active stack.
    stack_updated = Event

    # This is the name of the command that can be undone.  It will be empty if
    # there is no command that can be undone.  It is maintained by the undo
    # manager.
    undo_name = Property(Unicode)

    ###########################################################################
    # 'IUndoManager' interface.
    ###########################################################################

    def redo(self):
        """ Redo the last undone command of the active command stack. """

        if self.active_stack is not None:
            self.active_stack.redo()

    def undo(self):
        """ Undo the last command of the active command stack. """

        if self.active_stack is not None:
            self.active_stack.undo()

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _active_stack_changed(self, new):
        """ Handle a different stack becoming active. """

        # Pretend that the stack contents have changed.
        self.stack_updated = new

    def _get_active_stack_clean(self):
        """ Get the current clean state. """

        if self.active_stack is None:
            active_stack_clean = True
        else:
            active_stack_clean = self.active_stack.clean

        return active_stack_clean

    def _get_redo_name(self):
        """ Get the current redo name. """

        if self.active_stack is None:
            redo_name = ""
        else:
            redo_name = self.active_stack.redo_name

        return redo_name

    def _get_undo_name(self):
        """ Get the current undo name. """

        if self.active_stack is None:
            undo_name = ""
        else:
            undo_name = self.active_stack.undo_name

        return undo_name



from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..action import Action
from ..action_event import ActionEvent


class TestAction(unittest.TestCase):

    def test_default_id(self):
        action = Action(name='Test')
        self.assertEqual(action.id, 'Test')

    def test_id(self):
        action = Action(name='Test', id='test')
        self.assertEqual(action.id, 'test')

    def test_perform(self):
        # test whether function is called by updating list
        # XXX should really use mock
        memo = []

        def perform():
            memo.append('called')

        action = Action(name='Test', on_perform=perform)
        event = ActionEvent()
        action.perform(event)
        self.assertEqual(memo, ['called'])

    def test_perform_none(self):
        action = Action(name='Test')
        event = ActionEvent()
        # does nothing, but shouldn't error
        action.perform(event)

    def test_destroy(self):
        action = Action(name='Test')
        # does nothing, but shouldn't error
        action.destroy()



""" (Pyface) Action for running a help demo.

    :Copyright: 2009, Enthought Inc.
    :License: BSD
    :Author: Vibha Srinivasan
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

# Enthought library imports.
from doc_action import DocAction

# This module's parent package.
PARENT = '.'.join(__name__.split('.')[:-2])

class LoadURLAction(DocAction):
    """ Workbench Action for displaying a url in a web browser.
    """

    def _my_help_doc_default(self):
        exns = self.extension_registry.get_extensions(PARENT +
                                                      '.help_downloads')
        for hc in exns:
            if hc.label == self.name:
                return hc
        return None



""" Tests for safe weakrefs. """


# Standard library imports.
import weakref

# Enthought library imports.
from envisage.safeweakref import ref
from traits.api import HasTraits
from traits.testing.unittest_tools import unittest


class SafeWeakrefTestCase(unittest.TestCase):
    """ Tests for safe weakrefs. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_can_create_weakref_to_bound_method(self):
        class Foo(HasTraits):
            def method(self):
                self.method_called = True

        f = Foo()

        # Get a weak reference to a bound method.
        r = ref(f.method)
        self.assertNotEqual(None, r())

        # Make sure we can call it.
        r()()
        self.assert_(f.method_called)

        # Delete the object to delete the method!
        del f

        # The reference should now return None.
        self.assertEqual(None, r())

        return

    def test_two_weakrefs_to_bound_method_are_identical(self):
        class Foo(HasTraits):
            def method(self):
                pass

        f = Foo()

        self.assertIs(ref(f.method), ref(f.method))

        return

    def test_internal_cache_is_weak_too(self):
        # smell: Fragile test because we are reaching into the internals of the
        # object under test.
        #
        # I can't see a (clean!) way around this without adding something to
        # the public API that would only exist for testing, but in terms of
        # 'bang for the buck' I think this is good enough despite the
        # fragility.
        cache = ref._cache

        class Foo(HasTraits):
            def method(self):
                pass

        f = Foo()

        # Get the length of the cache before we do anything.
        len_cache = len(cache)

        # Create a weak reference to the bound method and make sure that
        # exactly one item has been added to the cache.
        r = ref(f.method)
        self.assertEqual(len_cache + 1, len(cache))

        # Delete the instance!
        del f

        # Our `ref` should now reference nothing...
        self.assertEqual(None, r())

        # ... and the cache should be back to its original size!
        self.assertEqual(len_cache, len(cache))

        return

    def test_two_weakrefs_to_bound_method_are_equal(self):
        class Foo(HasTraits):
            def method(self):
                pass

        f = Foo()

        # Make sure that two references to the same method compare as equal.
        r1 = ref(f.method)
        r2 = ref(f.method)
        self.assertEqual(r1, r2)

        # Make sure that a reference compares as unequal to non-references!
        self.assert_(not r1 == 99)

        return

    def test_two_weakrefs_to_bound_method_hash_equally(self):
        class Foo(HasTraits):
            def method(self):
                pass

        f = Foo()

        # Make sure we can hash the references.
        r1 = ref(f.method)
        r2 = ref(f.method)

        self.assertEqual(hash(r1), hash(r2))

        # Make sure we can hash non-bound methods.
        r1 = ref(Foo)
        r2 = ref(Foo)

        self.assertEqual(hash(r1), hash(r2))

        return

    def test_get_builtin_weakref_for_non_bound_method(self):
        class Foo(HasTraits):
            pass

        f = Foo()

        # Get a weak reference to something that is not a bound method.
        r = ref(f)
        self.assertEqual(weakref.ref, type(r))

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################
